from elevenlabs.client import ElevenLabs
from elevenlabs import Voice, VoiceSettings
from typing import Iterator, Union
from .base_engine import BaseEngine
import subprocess
import threading
import logging
import pyaudio


class ElevenlabsVoice:
    def __init__(self, name, voice_id, category, description, labels):
        self.name = name
        self.voice_id = voice_id
        self.category = category
        self.description = description
        self.labels = labels

    def __repr__(self):
        label_string = ", ".join(self.labels.values())
        return f"{self.name} ({self.category}, id: {self.voice_id}, {label_string})"


class ElevenlabsEngine(BaseEngine):
    def __init__(
        self,
        api_key: str = "",
        voice: str = "Nicole",
        id: str = "piTKgcLEGmPE4e6mEKli",
        category: str = "premade",
        clarity: float = 75.0,
        stability: float = 50.0,
        style_exxageration: float = 0.0,
        model: str = "eleven_multilingual_v1",
    ):
        """
        Initializes an elevenlabs voice realtime text to speech engine object.

        Args:
            api_key (str): Elevenlabs API key. (TTS API key)
            voice (str, optional): Voice name. Defaults to "Nicole".
            id (str, optional): Voice ID. Defaults to "piTKgcLEGmPE4e6mEKli".
            category (str, optional): Voice category. Defaults to "premade".
            clarity (float, optional): Clarity / Similarity. Adjusts voice similarity and resemblance. Defaults to "75.0".
            stability (float, optional): Stability. Controls the voice performance, with higher values producing a steadier tone and lower values giving a more emotive output. Defaults to "50.0".
            style_exxageration (float, optional): Style Exxageration. Controls the voice performance, with higher values giving a more emotive output and lower values producing a steadier tone. Defaults to "0.0".
            model (str, optional): Model. Defaults to "eleven_multilingual_v1". Some models may not work with real time inference.
        """

        self.voice_name = voice
        self.id = id
        self.category = category
        self.clarity = clarity
        self.stability = stability
        self.style_exxageration = style_exxageration
        self.model = model
        self.pause_event = threading.Event()
        self.immediate_stop = threading.Event()
        self.on_audio_chunk = None
        self.muted = False
        self.on_playback_started = False

        self.client = ElevenLabs(api_key=api_key)

    def post_init(self):
        """Information that this engine can handle generators directly"""
        self.can_consume_generators = True
        self.engine_name = "elevenlabs"

    def get_stream_info(self):
        """
        Returns the audio stream configuration information suitable for PyAudio.

        Returns:
            tuple: A tuple containing the audio format, number of channels, and the sample rate.
                  - Format (int): The format of the audio stream. pyaudio.paInt16 represents 16-bit integers.
                  - Channels (int): The number of audio channels. 1 represents mono audio.
                  - Sample Rate (int): The sample rate of the audio in Hz. 16000 represents 16kHz sample rate.
        """
        return pyaudio.paCustomFormat, -1, -1

    def pause(self):
        """Pauses playback of the synthesized audio stream (won't work properly with elevenlabs)."""
        self.pause_event.set()

    def set_muted(self, muted: bool):
        """Mutes the audio stream."""
        self.muted = muted

    def resume(self):
        """Resumes a previously paused playback of the synthesized audio stream (won't work properly with elevenlabs)."""
        self.pause_event.clear()

    def stop(self):
        """Stops the playback of the synthesized audio stream immediately."""
        self.mpv_process.terminate()
        return True

    def synthesize(self, generator: Iterator[str]) -> bool:
        """
        Synthesizes text to audio stream.

        Args:
            text (str): Text to synthesize.
        """

        self.on_playback_started = False
        self.immediate_stop.clear()

        voice = Voice(
            voice_id=self.id,
            settings=VoiceSettings(
                stability=self.stability / 100,
                similarity_boost=self.clarity / 100,
                style=self.style_exxageration / 100,
                use_speaker_boost=True,
            ),
        )

        self.audio_stream = self.client.generate(
            text=generator, voice=voice, model=self.model, stream=True
        )

        self.stream(self.audio_stream)

        return True

    def set_api_key(self, api_key: str):
        """
        Sets the elevenlabs api key.

        Args:
            api_key (str): Elevenlabs API key. (TTS API key)
        """
        self.api_key = api_key
        if api_key:
            self.client = ElevenLabs(api_key=api_key)

    def stream(self, audio_stream: Iterator[bytes]) -> bytes:
        """
        Stream the audio data using the 'mpv' player.

        This method takes the audio_stream iterator, which contains bytes of audio data,
        and plays them using the 'mpv' player. The function will continuously feed the
        'mpv' process with chunks of audio data until the audio is finished or an error
        occurs.

        If the 'mpv' player is not installed, a ValueError is raised.

        Args:
            audio_stream (Iterator[bytes]): An iterator that yields bytes of audio data.
            filename (Optional[str], optional): The filename to save the audio to. Defaults to None.

        Returns:
            bytes: The concatenated bytes of the entire audio stream.

        Raises:
            ValueError: If the 'mpv' player is not found.
        """
        if not self.is_installed("mpv"):
            message = (
                "mpv not found, necessary to stream audio. "
                "On mac you can install it with 'brew install mpv'. "
                "On linux and windows you can install it from https://mpv.io/"
            )
            raise ValueError(message)

        if not self.muted:
            mpv_command = ["mpv", "--no-cache", "--no-terminal", "--", "fd://0"]
            self.mpv_process = subprocess.Popen(
                mpv_command,
                stdin=subprocess.PIPE,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )

        audio = b""

        try:
            for chunk in audio_stream:
                if chunk is not None:
                    if not self.on_playback_started and self.on_playback_start:
                        self.on_playback_start()
                    self.on_playback_started = True
                    if not self.muted:
                        self.mpv_process.stdin.write(chunk)
                        self.mpv_process.stdin.flush()
                    audio += chunk
                    if self.on_audio_chunk:
                        self.on_audio_chunk(chunk)
        except BrokenPipeError:
            pass

        if not self.muted:
            if self.mpv_process.stdin:
                self.mpv_process.stdin.close()
            self.mpv_process.wait()

        return audio

    def get_voices(self):
        """
        Retrieves the voices available from the Elevenlabs voice source.

        Calling this takes time, it sends a request to the Elevenlabs API to fetch the list of available voices.
        This method fetches the list of available voices using the elevenlabs `voices()` function and then
        constructs a list of `ElevenlabsVoice` objects to represent each voice's details.

        Returns:
            list[ElevenlabsVoice]: A list containing ElevenlabsVoice objects representing each available voice.
                                Each ElevenlabsVoice object encapsulates information such as the voice's name,
                                ID, category, description, and associated labels.

        Note:
            This method relies on the `voices()` function to obtain the raw voice data. Ensure that the
            `voices()` function is accessible and functional before calling this method.
        """
        fetched_voices = self.client.voices.get_all()

        voices_list = fetched_voices.voices

        voice_objects = []
        for voice in voices_list:
            voice_object = ElevenlabsVoice(
                name=voice.name,
                voice_id=voice.voice_id,
                category=voice.category,
                description=voice.description,
                labels=voice.labels,
            )
            voice_objects.append(voice_object)

        return voice_objects

    def set_voice(self, voice: Union[str, ElevenlabsVoice]):
        """
        Sets the voice to be used for speech synthesis.

        Args:
            voice (Union[str, ElevenlabsVoice]): The voice to be used for speech synthesis.
        """
        if isinstance(voice, ElevenlabsVoice):
            logging.info(f"Setting voice to {voice.name}")
            self.voice_name = voice.name
            self.id = voice.voice_id
            self.category = voice.category
            return
        else:
            installed_voices = self.get_voices()
            for installed_voice in installed_voices:
                if voice in installed_voice.name:
                    logging.info(f"Setting voice to {installed_voice.name}")
                    self.voice_name = installed_voice.name
                    self.id = installed_voice.voice_id
                    self.category = installed_voice.category
                    return

        logging.warning(f"Voice {voice} not found.")

    def set_voice_parameters(self, **voice_parameters):
        """
        Sets the voice parameters to be used for speech synthesis.

        Args:
            **voice_parameters: The voice parameters to be used for speech synthesis.
        """
        if "clarity" in voice_parameters:
            self.clarity = voice_parameters["clarity"]
        if "stability" in voice_parameters:
            self.stability = voice_parameters["stability"]
        if "style_exxageration" in voice_parameters:
            self.style_exxageration = voice_parameters["style_exxageration"]



import azure.cognitiveservices.speech as tts
from .base_engine import BaseEngine
from typing import Union
import requests
import pyaudio
import logging


class PushAudioOutputStreamSampleCallback(tts.audio.PushAudioOutputStreamCallback):
    """
    This class provides a callback mechanism to handle audio output streams for Azure's Text-to-Speech (TTS) service.
    It allows you to capture synthesized audio data in real-time and push it to a buffer.

    Attributes:
        buffer: A buffer or queue where the audio stream data will be stored.
    """

    def __init__(self, buffer):
        self.buffer = buffer

    def write(self, audio_buffer: memoryview) -> int:
        self.buffer.put(audio_buffer.tobytes())
        return audio_buffer.nbytes


class AzureVoice:
    def __init__(self, name, locale, gender):
        self.full_name = name
        self.name = AzureVoice._extract_voice_name(self.full_name)
        self.locale = locale
        self.language = AzureVoice._extract_voice_language(self.locale)
        self.gender = gender

    def __repr__(self):
        return f"{self.name} ({self.gender}, {self.locale})"

    @staticmethod
    def _extract_voice_name(full_name):
        # Extracts the real voice name from the full "Name" string
        start_index = full_name.rfind(", ") + 2
        end_index = full_name.rfind(")")
        return full_name[start_index:end_index]

    @staticmethod
    def _extract_voice_language(locale):
        # Extracts the language from the locale string
        end_index = locale.find("-")
        return locale[:end_index]


class AzureEngine(BaseEngine):
    def __init__(
        self,
        speech_key: str = "",
        service_region: str = "",
        voice: str = "en-US-AshleyNeural",
        rate: float = 0.0,
        pitch: float = 0.0,
    ):
        """
        Initializes an azure voice realtime text to speech engine object.

        Args:
            speech_key (str): Azure subscription key. (TTS API key)
            service_region (str): Azure service region. (Cloud Region ID)
            voice (str, optional): Voice name. Defaults to "en-US-AshleyNeural".
            rate (float, optional): Speech speed as a percentage. Defaults to "0.0". Indicating the relative change.
            pitch (float, optional): Speech pitch as a percentage. Defaults to "0.0". Indicating the relative change.
        """

        self.speech_key = speech_key
        self.service_region = service_region
        self.language = voice[:5]
        self.voice_name = voice
        self.rate = rate
        self.pitch = pitch
        self.emotion = "neutral"
        self.emotion_degree = 1.0
        self.emotion_role = "YoungAdultFemale"
        self.emotion_roles = [
            "Girl",
            "Boy",
            "YoungAdultFemale",
            "YoungAdultMale",
            "OlderAdultFemale",
            "OlderAdultMale",
            "SeniorFemale",
            "SeniorMale",
        ]
        self.emotions = [
            "advertisement_upbeat",
            "affectionate",
            "angry",
            "assistant",
            "calm",
            "chat",
            "cheerful",
            "customerservice",
            "depressed",
            "disgruntled",
            "documentary-narration",
            "embarrassed",
            "empathetic",
            "envious",
            "excited",
            "fearful",
            "friendly",
            "gentle",
            "hopeful",
            "lyrical",
            "narration-professional",
            "narration-relaxed",
            "neutral",
            "newscast",
            "newscast-casual",
            "newscast-formal",
            "poetry-reading",
            "sad",
            "serious",
            "shouting",
            "sports_commentary",
            "sports_commentary_excited",
            "whispering",
            "terrified",
            "unfriendly",
        ]

    def post_init(self):
        self.engine_name = "azure"

    def get_stream_info(self):
        """
        Returns the PyAudio stream configuration information suitable for Azure Engine.

        Returns:
            tuple: A tuple containing the audio format, number of channels, and the sample rate.
                  - Format (int): The format of the audio stream. pyaudio.paInt16 represents 16-bit integers.
                  - Channels (int): The number of audio channels. 1 represents mono audio.
                  - Sample Rate (int): The sample rate of the audio in Hz. 16000 represents 16kHz sample rate.
        """
        return pyaudio.paInt16, 1, 16000

    def synthesize(self, text: str) -> bool:
        """
        Synthesizes text to audio stream.

        Args:
            text (str): Text to synthesize.
        """

        # Set up the Azure TTS configuration
        speech_config = tts.SpeechConfig(
            subscription=self.speech_key, region=self.service_region
        )
        stream_callback = PushAudioOutputStreamSampleCallback(self.queue)
        push_stream = tts.audio.PushAudioOutputStream(stream_callback)
        stream_config = tts.audio.AudioOutputConfig(stream=push_stream)
        speech_synthesizer = tts.SpeechSynthesizer(
            speech_config=speech_config, audio_config=stream_config
        )

        emotion_start_tag = f'<mstts:express-as style="{self.emotion}" styledegree="{self.emotion_degree}" role="{self.emotion_role}">'
        emotion_end_tag = "</mstts:express-as>"
        if self.emotion not in self.emotions or self.emotion == "neutral":
            emotion_start_tag = ""
            emotion_end_tag = ""

        ssml_string = f"""
        <speak version="1.0" xmlns="http://www.w3.org/2001/10/synthesis" xmlns:mstts="https://www.w3.org/2001/mstts" xml:lang="{self.language}">
            <voice name="{self.voice_name}">
                {emotion_start_tag}
                    <prosody rate="{self.rate}%" pitch="{self.pitch}%">
                        {text}
                    </prosody>
                {emotion_end_tag}
            </voice>
        </speak>
        """

        logging.debug(f"SSML:\n{ssml_string}")

        # Convert the SSML to audio stream
        result = speech_synthesizer.speak_ssml_async(ssml_string).get()

        if result.reason == tts.ResultReason.SynthesizingAudioCompleted:
            logging.debug("Speech synthesized")
            return True
        elif result.reason == tts.ResultReason.Canceled:
            cancellation_details = result.cancellation_details
            print(
                f"Speech synthesis canceled, check speech_key and service_region: {result.reason}"
            )
            print("Cancellation details: {}".format(cancellation_details.reason))
            print("SSLM:")
            print(ssml_string)
            if cancellation_details.reason == tts.CancellationReason.Error:
                print("Error details: {}".format(cancellation_details.error_details))
        else:
            print(f"Speech synthesis failed: {result.reason}")
            print(f"Result: {result}")

    def set_emotion(
        self,
        emotion: str,
        emotion_role: str = "YoungAdultFemale",
        emotion_degree: float = 1.0,
    ):
        """
        Sets the emotion to be used for speech synthesis.

        Args:
            emotion (str): The emotion to be used for speech synthesis.
            emotion_degree (float, optional): The degree of the emotion. Defaults to 1.0.
            emotion_role (str, optional): The role of the emotion. Defaults to "YoungAdultFemale".
        """
        # https://learn.microsoft.com/en-us/azure/ai-services/speech-service/speech-synthesis-markup-voice
        self.emotion = emotion
        self.emotion_degree = emotion_degree
        self.emotion_role = emotion_role

    def get_emotions(self):
        """
        Retrieves the available emotions for the Azure Speech engine.

        Returns:
            list[str]: A list containing the available emotions for the Azure Speech engine.
        """
        return self.emotions

    def set_speech_key(self, speech_key: str):
        """
        Sets the azure subscription key.

        Args:
            speech_key (str): Azure subscription key. (TTS API key)
        """
        self.speech_key = speech_key

    def set_service_region(self, service_region: str):
        """
        Sets the azure service region.

        Args:
            service_region (str): Azure service region. (Cloud Region ID)
        """
        self.service_region = service_region

    def get_voices(self):
        """
        Retrieves the installed voices available for the Azure Speech engine.

        Sends a request to the Azure Speech API to fetch the list of available voices.
        The method uses the `service_region` and `speech_key` attributes of the instance to authenticate
        and get the list of voices.

        Returns:
            list[AzureVoice]: A list containing AzureVoice objects representing the available voices.
                            Each AzureVoice object encapsulates information like the real name, locale,
                            and gender of the voice. If the API call fails, an empty list is returned.

        Raises:
            May raise exceptions related to the `requests` module like ConnectionError, Timeout, etc.

        Side Effects:
            Makes HTTP requests to the Azure Speech API. Prints an error message to stdout if the
            request fails.

        Note:
            Ensure that `self.service_region` and `self.speech_key` are correctly set before calling
            this method.
        """
        token_endpoint = f"https://{self.service_region}.api.cognitive.microsoft.com/sts/v1.0/issueToken"
        headers = {"Ocp-Apim-Subscription-Key": self.speech_key}
        response = requests.post(token_endpoint, headers=headers)
        access_token = str(response.text)

        fetch_voices_endpoint = f"https://{self.service_region}.tts.speech.microsoft.com/cognitiveservices/voices/list"
        voice_headers = {"Authorization": "Bearer " + access_token}
        response = requests.get(fetch_voices_endpoint, headers=voice_headers)

        voice_objects = []

        if response.status_code == 200:
            voices = response.json()
            for voice in voices:
                real_name = voice["Name"]
                locale = voice["Locale"]
                gender = voice.get("Gender", "N/A")
                voice_object = AzureVoice(real_name, locale, gender)
                voice_objects.append(voice_object)

            return voice_objects
        else:
            print(f"Error {response.status_code}: {response.text}")
            return []

    def set_voice(self, voice: Union[str, AzureVoice]):
        """
        Sets the voice to be used for speech synthesis.

        Args:
            voice (Union[str, AzureVoice]): The voice to be used for speech synthesis.
        """
        if isinstance(voice, AzureVoice):
            self.voice_name = voice.full_name
            self.language = self.voice_name[:5]
        else:
            installed_voices = self.get_voices()
            for installed_voice in installed_voices:
                if voice in installed_voice.full_name:
                    self.voice_name = installed_voice.full_name
                    self.language = self.voice_name[:5]

    def set_voice_parameters(self, **voice_parameters):
        """
        Sets the voice parameters to be used for speech synthesis.

        Args:
            **voice_parameters: The voice parameters to be used for speech synthesis.
        """
        if "rate" in voice_parameters:
            self.rate = voice_parameters["rate"]
        if "pitch" in voice_parameters:
            self.pitch = voice_parameters["pitch"]



"""
Stream management
"""

from pydub import AudioSegment
import threading
import pyaudio
import logging
import queue
import time
import io


class AudioConfiguration:
    """
    Defines the configuration for an audio stream.
    """

    def __init__(
        self,
        format: int = pyaudio.paInt16,
        channels: int = 1,
        rate: int = 16000,
        output_device_index=None,
        muted: bool = False,
    ):
        """
        Args:
            format (int): Audio format, defaults to pyaudio.paInt16
            channels (int): Number of channels, defaults to 1 (mono)
            rate (int): Sample rate, defaults to 16000
            output_device_index (int): Output device index, defaults to None
        """
        self.format = format
        self.channels = channels
        self.rate = rate
        self.output_device_index = output_device_index
        self.muted = muted


class AudioStream:
    """
    Handles audio stream operations
    - opening, starting, stopping, and closing
    """

    def __init__(self, config: AudioConfiguration):
        """
        Args:
            config (AudioConfiguration): Object containing audio settings.
        """
        self.config = config
        self.stream = None
        self.pyaudio_instance = pyaudio.PyAudio()

    def open_stream(self):
        """Opens an audio stream."""

        # check for mpeg format
        pyChannels = self.config.channels
        pySampleRate = self.config.rate
        pyOutput_device_index = self.config.output_device_index

        if self.config.muted:
            logging.debug("Muted mode, no opening stream")

        else:
            if self.config.format == pyaudio.paCustomFormat:
                pyFormat = self.pyaudio_instance.get_format_from_width(2)
                logging.debug(
                    "Opening stream for mpeg audio chunks, "
                    f"pyFormat: {pyFormat}, pyChannels: {pyChannels}, "
                    f"pySampleRate: {pySampleRate}"
                )
            else:
                pyFormat = self.config.format
                logging.debug(
                    "Opening stream for wave audio chunks, "
                    f"pyFormat: {pyFormat}, pyChannels: {pyChannels}, "
                    f"pySampleRate: {pySampleRate}"
                )

            try:
                self.stream = self.pyaudio_instance.open(
                    format=pyFormat,
                    channels=pyChannels,
                    rate=pySampleRate,
                    output_device_index=pyOutput_device_index,
                    output=True,
                )
            except Exception as e:
                print(f"Error opening stream: {e}")
                exit(0)

    def start_stream(self):
        """Starts the audio stream."""
        if self.stream and not self.stream.is_active():
            self.stream.start_stream()

    def stop_stream(self):
        """Stops the audio stream."""
        if self.stream and self.stream.is_active():
            self.stream.stop_stream()

    def close_stream(self):
        """Closes the audio stream."""
        if self.stream:
            self.stop_stream()
            self.stream.close()
            self.stream = None

    def is_stream_active(self) -> bool:
        """
        Checks if the audio stream is active.

        Returns:
            bool: True if the stream is active, False otherwise.
        """
        return self.stream and self.stream.is_active()


class AudioBufferManager:
    """
    Manages an audio buffer, allowing addition and retrieval of audio data.
    """

    def __init__(self, audio_buffer: queue.Queue):
        """
        Args:
            audio_buffer (queue.Queue): Queue to be used as the audio buffer.
        """
        self.audio_buffer = audio_buffer
        self.total_samples = 0

    def add_to_buffer(self, audio_data):
        """
        Adds audio data to the buffer.

        Args:
            audio_data: Audio data to be added.
        """
        self.audio_buffer.put(audio_data)
        self.total_samples += len(audio_data) // 2

    def clear_buffer(self):
        """Clears all audio data from the buffer."""
        while not self.audio_buffer.empty():
            try:
                self.audio_buffer.get_nowait()
            except queue.Empty:
                continue
        self.total_samples = 0

    def get_from_buffer(self, timeout: float = 0.05):
        """
        Retrieves audio data from the buffer.

        Args:
            timeout (float): Time (in seconds) to wait
              before raising a queue.Empty exception.

        Returns:
            The audio data chunk or None if the buffer is empty.
        """
        try:
            chunk = self.audio_buffer.get(timeout=timeout)
            self.total_samples -= len(chunk) // 2
            return chunk
        except queue.Empty:
            return None

    def get_buffered_seconds(self, rate: int) -> float:
        """
        Calculates the duration (in seconds) of the buffered audio data.

        Args:
            rate (int): Sample rate of the audio data.

        Returns:
            float: Duration of buffered audio in seconds.
        """
        return self.total_samples / rate


class StreamPlayer:
    """
    Manages audio playback operations such as start, stop, pause, and resume.
    """

    def __init__(
        self,
        audio_buffer: queue.Queue,
        config: AudioConfiguration,
        on_playback_start=None,
        on_playback_stop=None,
        on_audio_chunk=None,
        muted=False,
    ):
        """
        Args:
            audio_buffer (queue.Queue): Queue to be used as the audio buffer.
            config (AudioConfiguration): Object containing audio settings.
            on_playback_start (Callable, optional): Callback function to be
              called at the start of playback. Defaults to None.
            on_playback_stop (Callable, optional): Callback function to be
              called at the stop of playback. Defaults to None.
        """
        self.buffer_manager = AudioBufferManager(audio_buffer)
        self.audio_stream = AudioStream(config)
        self.playback_active = False
        self.immediate_stop = threading.Event()
        self.pause_event = threading.Event()
        self.playback_thread = None
        self.on_playback_start = on_playback_start
        self.on_playback_stop = on_playback_stop
        self.on_audio_chunk = on_audio_chunk
        self.first_chunk_played = False
        self.muted = muted

    def _play_chunk(self, chunk):
        """
        Plays a chunk of audio data.

        Args:
            chunk: Chunk of audio data to be played.
        """

        # handle mpeg
        if self.audio_stream.config.format == pyaudio.paCustomFormat:
            # convert to pcm using pydub
            segment = AudioSegment.from_file(io.BytesIO(chunk), format="mp3")
            chunk = segment.raw_data

        sub_chunk_size = 1024

        for i in range(0, len(chunk), sub_chunk_size):
            sub_chunk = chunk[i : i + sub_chunk_size]

            if not self.first_chunk_played and self.on_playback_start:
                self.on_playback_start()
                self.first_chunk_played = True

            if not self.muted:
                self.audio_stream.stream.write(sub_chunk)

            if self.on_audio_chunk:
                self.on_audio_chunk(sub_chunk)

            # Pause playback if the event is set
            while self.pause_event.is_set():
                time.sleep(0.01)

            if self.immediate_stop.is_set():
                break

    def _process_buffer(self):
        """
        Processes and plays audio data from the buffer
        until it's empty or playback is stopped.
        """
        while self.playback_active or not self.buffer_manager.audio_buffer.empty():
            chunk = self.buffer_manager.get_from_buffer()
            if chunk:
                self._play_chunk(chunk)

            if self.immediate_stop.is_set():
                logging.info("Immediate stop requested, aborting playback")
                break
        if self.on_playback_stop:
            self.on_playback_stop()

    def get_buffered_seconds(self) -> float:
        """
        Calculates the duration (in seconds) of the buffered audio data.

        Returns:
            float: Duration of buffered audio in seconds.
        """
        total_samples = sum(
            len(chunk) // 2 for chunk in list(self.buffer_manager.audio_buffer.queue)
        )
        return total_samples / self.audio_stream.config.rate

    def start(self):
        """Starts audio playback."""
        self.first_chunk_played = False
        self.playback_active = True
        if not self.audio_stream.stream:
            self.audio_stream.open_stream()

        self.audio_stream.start_stream()

        if not self.playback_thread or not self.playback_thread.is_alive():
            self.playback_thread = threading.Thread(target=self._process_buffer)
            self.playback_thread.start()

    def stop(self, immediate: bool = False):
        """
        Stops audio playback.

        Args:
            immediate (bool): If True, stops playback immediately
              without waiting for buffer to empty.
        """
        if not self.playback_thread:
            logging.warn("No playback thread found, cannot stop playback")
            return

        if immediate:
            self.immediate_stop.set()
            while self.playback_active:
                time.sleep(0.1)
            return

        self.playback_active = False

        if self.playback_thread and self.playback_thread.is_alive():
            self.playback_thread.join()

        time.sleep(0.1)

        self.audio_stream.close_stream()
        self.immediate_stop.clear()
        self.buffer_manager.clear_buffer()
        self.playback_thread = None

    def pause(self):
        """Pauses audio playback."""
        self.pause_event.set()

    def resume(self):
        """Resumes paused audio playback."""
        self.pause_event.clear()

    def mute(self, muted: bool = True):
        """Mutes audio playback."""
        self.muted = muted



from .base_engine import BaseEngine
import torch.multiprocessing as mp
from threading import Lock, Thread
from typing import Union, List
from pathlib import Path
from tqdm import tqdm
import numpy as np
import traceback
import requests
import logging
import pyaudio
import torch
import time
import json
import sys
import io
import os
import re


TIME_SLEEP_DEVICE_RESET = 2


class QueueWriter(io.TextIOBase):
    """
    Custom file-like object to write text to a multiprocessing queue.
    """

    def __init__(self, queue):
        super().__init__()
        self.queue = queue

    def write(self, msg):
        """
        Write the message to the queue.

        Args:
            msg (str): The message to write.
        """
        if msg.strip():  # Avoid sending empty strings and newline characters.
            self.queue.put(msg)


class CoquiVoice:
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return f"{self.name}"


class CoquiEngine(BaseEngine):
    def __init__(
        self,
        model_name="tts_models/multilingual/multi-dataset/xtts_v2",
        specific_model="v2.0.2",
        local_models_path=None,
        voices_path=None,
        voice: Union[str, List[str]] = "",
        language="en",
        speed=1.0,
        thread_count=6,
        stream_chunk_size=20,
        overlap_wav_len=1024,
        temperature=0.85,
        length_penalty=1.0,
        repetition_penalty=7.0,
        top_k=50,
        top_p=0.85,
        enable_text_splitting=True,
        full_sentences=False,
        level=logging.WARNING,
        use_deepspeed=False,
        device: str = None,
        prepare_text_for_synthesis_callback=None,
        add_sentence_filter=False,
        pretrained=False,
        comma_silence_duration=0.3,
        sentence_silence_duration=0.6,
        default_silence_duration=0.3,
        print_realtime_factor=False,
        load_balancing=False,
        load_balancing_buffer_length=0,
        load_balancing_cut_off=0,
    ):
        """
        Initializes a coqui voice realtime text to speech engine object.

        Args:
            model_name (str):
              Name of the coqui model to use.
              Tested with xtts_v2 only.
            specific_model (str):
              Name of the specific model to use.
              If not specified, the most recent model will be downloaded.
            local_models_path (str):
              Path to a local models directory.
              If not specified, a directory "models" will be created in the
              script directory.
            voice (str):
              Name to the file containing the voice to clone.
              Works with a 44100Hz or 22050Hz mono 32bit float WAV file.
            language (str):
              Language to use for the coqui model.
            speed (float):
              Speed factor for the coqui model.
            thread_count (int):
              Number of threads to use for the coqui model.
            stream_chunk_size (int):
              Chunk size for the coqui model.
            overlap_wav_len (int):
              Overlap length for the coqui model.
            temperature (float):
              Temperature for the coqui model.
            length_penalty (float):
              Length penalty for the coqui model.
            repetition_penalty (float):
              Repetition penalty for the model.
            top_k (int):
              Top K for the coqui model.
            top_p (float):
              Top P for the coqui model.
            enable_text_splitting (bool):
              Enable text splitting for the model.
            full_sentences (bool):
              Enable full sentences for the coqui model.
            level (int):
              Logging level for the coqui model.
            use_deepspeed (bool):
              Enable deepspeed for the coqui model.
            device (str):
              Specify the device to use for model inference ("cuda", "mps", "cpu").
              If not specified or invalid, the device will be automatically selected.
            prepare_text_for_synthesis_callback (function):
              Function to prepare text for synthesis.
              If not specified, a default sentence parser will be used.
            add_sentence_filter (bool):
              Adds a own sentence filter in addition
              to the one coqui tts already provides.
            pretrained (bool):
              Use a pretrained model for the coqui model.
            comma_silence_duration (float):
              Duration of the silence after a comma.
            sentence_silence_duration (float):
              Duration of the silence after a sentence.
            default_silence_duration (float):
                Default duration of the silence.
            print_realtime_factor (bool):
                Print the realtime factor for the coqui model.
            load_balancing (bool):
                Enable load balancing for the coqui model.
            load_balancing_buffer_length (int):
                Buffer length for the load balancing.
            load_balancing_cut_off (int):
                Cut off for the load balancing.
        """

        self._synthesize_lock = Lock()
        self.model_name = model_name
        self.pretrained = pretrained
        self.language = language
        self.level = level
        self.thread_count = thread_count
        self.stream_chunk_size = stream_chunk_size
        self.overlap_wav_len = overlap_wav_len
        self.temperature = temperature
        self.length_penalty = length_penalty
        self.repetition_penalty = repetition_penalty
        self.top_k = top_k
        self.top_p = top_p
        self.enable_text_splitting = enable_text_splitting
        self.full_sentences = full_sentences
        self.use_deepspeed = use_deepspeed
        self.device = device
        self.add_sentence_filter = add_sentence_filter
        self.comma_silence_duration = comma_silence_duration
        self.sentence_silence_duration = sentence_silence_duration
        self.default_silence_duration = default_silence_duration
        self.print_realtime_factor = print_realtime_factor
        self.load_balancing = load_balancing
        self.load_balancing_buffer_length = load_balancing_buffer_length
        self.load_balancing_cut_off = load_balancing_cut_off

        self.cloning_reference_wav = voice
        self.speed = speed
        self.specific_model = specific_model
        if not local_models_path:
            local_models_path = os.environ.get("COQUI_MODEL_PATH")
            if local_models_path and len(local_models_path) > 0:
                logging.info(
                    "Local models path from environment variable "
                    f'COQUI_MODEL_PATH: "{local_models_path}"'
                )
        self.local_models_path = local_models_path
        self.prepare_text_callback = prepare_text_for_synthesis_callback

        self.voices_path = voices_path

        # download coqui model
        self.model_path = None
        if not self.specific_model:
            from TTS.utils.manage import ModelManager

            logging.info("Download most recent XTTS Model if available")
            ModelManager().download_model(model_name)
        else:
            logging.info(f'Local XTTS Model: "{specific_model}" specified')
            self.model_path = self.download_model(
                specific_model, self.local_models_path
            )

        # Start the worker process
        try:
            # Only set the start method if it hasn't been set already
            if mp.get_start_method(allow_none=True) is None:
                mp.set_start_method("spawn")
        except RuntimeError as e:
            print("Start method has already been set. Details:", e)

        self.create_worker_process()

    def create_worker_process(self):
        self.output_queue = mp.Queue()

        def output_worker(queue):
            """
            Worker function that prints messages from the queue.

            Args:
                queue (multiprocessing.Queue): Queue to receive messages from.
            """
            while True:
                message = queue.get()
                if message == "STOP":  # A special message to stop the worker.
                    break
                print(message)

        self.output_worker_thread = Thread(
            target=output_worker, args=(self.output_queue,)
        )
        self.output_worker_thread.daemon = True
        self.output_worker_thread.start()

        self.main_synthesize_ready_event = mp.Event()
        self.parent_synthesize_pipe, child_synthesize_pipe = mp.Pipe()
        self.voices_list = []
        self.retrieve_coqui_voices()

        self.synthesize_process = mp.Process(
            target=CoquiEngine._synthesize_worker,
            args=(
                self.output_queue,
                child_synthesize_pipe,
                self.model_name,
                self.cloning_reference_wav,
                self.language,
                self.main_synthesize_ready_event,
                self.level,
                self.speed,
                self.thread_count,
                self.stream_chunk_size,
                self.full_sentences,
                self.overlap_wav_len,
                self.temperature,
                self.length_penalty,
                self.repetition_penalty,
                self.top_k,
                self.top_p,
                self.enable_text_splitting,
                self.model_path,
                self.use_deepspeed,
                self.device,
                self.voices_path,
                self.voices_list,
                self.pretrained,
                self.comma_silence_duration,
                self.sentence_silence_duration,
                self.default_silence_duration,
                self.print_realtime_factor,
                self.load_balancing,
                self.load_balancing_buffer_length,
                self.load_balancing_cut_off,
            ),
        )
        self.synthesize_process.start()

        logging.debug("Waiting for coqui model start")
        self.main_synthesize_ready_event.wait()
        logging.info("Coqui synthesis model ready")

    def post_init(self):
        self.engine_name = "coqui"

    @staticmethod
    def _synthesize_worker(
        output_queue,
        conn,
        model_name,
        cloning_reference_wav: Union[str, List[str]],
        language,
        ready_event,
        loglevel,
        speed,
        thread_count,
        stream_chunk_size,
        full_sentences,
        overlap_wav_len,
        temperature,
        length_penalty,
        repetition_penalty,
        top_k,
        top_p,
        enable_text_splitting,
        local_model_path,
        use_deepspeed,
        device,
        voices_path,
        predefined_voices,
        pretrained,
        comma_silence_duration,
        sentence_silence_duration,
        default_silence_duration,
        print_realtime_factor,
        load_balancing,
        load_balancing_buffer_length,
        load_balancing_cut_off,
    ):
        """
        Worker process for the coqui text to speech synthesis model.

        Args:
            conn (multiprocessing.Connection):
              Connection to the parent process.
            model_name (str): Name of the coqui model to use.
            cloning_reference_wav (str):
              Name to the file containing the voice to clone.
            language (str): Language to use for the coqui model.
            ready_event (multiprocessing.Event):
              Event to signal when the model is ready.
        """
        sys.stdout = QueueWriter(output_queue)
        sys.stderr = QueueWriter(output_queue)

        from TTS.config import load_config
        from TTS.tts.models import setup_model as setup_tts_model
        from TTS.tts.layers.xtts.xtts_manager import SpeakerManager

        tts = None

        logging.basicConfig(format="CoquiEngine: %(message)s", level=loglevel)

        logging.info("Starting CoquiEngine")

        def get_conditioning_latents(filenames: Union[str, List[str]], tts):
            """
            Method still needs some rework
            """
            if not isinstance(filenames, list):
                filenames = [filenames]

            logging.debug("Computing speaker latents")

            if len(filenames) == 0 or not filenames[0]:
                logging.debug("Using coqui_default_voice.wav as default voice")
                filenames = ["coqui_default_voice.wav"]

            if len(filenames) == 1:
                logging.debug("Handling of single voice file")

                # verify that filename ends with .wav
                filename = filenames[0]
                if filename.endswith(".json"):
                    filename_json = filename
                    filename = filename[:-5]
                    filename_wav = filename + "wav"
                elif filename.endswith(".wav"):
                    filename_json = filename[:-3] + "json"
                    filename = filename[:-3]
                    filename_wav = filename + "wav"
                else:
                    filename_json = filename + ".json"
                    filename_wav = filename + ".wav"

                if voices_path:
                    filename_voice_wav = os.path.join(voices_path, filename_wav)
                    filename_voice_json = os.path.join(voices_path, filename_json)
                else:
                    filename_voice_wav = filename_wav
                    filename_voice_json = filename_json

                if not os.path.exists(filename_voice_json) and not os.path.exists(
                    filename_voice_wav
                ):
                    for predefined_voice in predefined_voices:
                        if filename.lower() in predefined_voice.lower():
                            speaker_file_path = os.path.join(
                                checkpoint, "speakers_xtts.pth"
                            )
                            speaker_manager = SpeakerManager(speaker_file_path)
                            gpt_cond_latent, speaker_embedding = (
                                speaker_manager.speakers[predefined_voice].values()
                            )
                            return gpt_cond_latent, speaker_embedding

                    if len(filename) > 0:
                        logging.info(
                            f"Using default voice, both {filename_voice_json} and {filename_voice_wav} not found."
                        )
                    else:
                        logging.info(
                            "Using default voice, no cloning source specified."
                        )

                    # Get the directory of the current script
                    current_dir = os.path.dirname(os.path.realpath(__file__))
                    filename_voice_json = os.path.join(
                        current_dir, "coqui_default_voice.json"
                    )
                    if not os.path.exists(filename_voice_json):
                        raise ValueError(
                            f"Default voice file {filename_voice_json} not found."
                        )

                # check if latents are already computed
                if os.path.exists(filename_voice_json):
                    logging.debug(
                        f"Latents already computed, reading from {filename_voice_json}"
                    )
                    with open(filename_voice_json, "r") as new_file:
                        latents = json.load(new_file)

                    speaker_embedding = (
                        torch.tensor(latents["speaker_embedding"])
                        .unsqueeze(0)
                        .unsqueeze(-1)
                    )
                    gpt_cond_latent = (
                        torch.tensor(latents["gpt_cond_latent"])
                        .reshape((-1, 1024))
                        .unsqueeze(0)
                    )

                    return gpt_cond_latent, speaker_embedding

                # compute and write latents to json file
                logging.debug(f"Computing latents for {filename}")

                gpt_cond_latent, speaker_embedding = tts.get_conditioning_latents(
                    audio_path=filename_voice_wav, gpt_cond_len=30, max_ref_length=60
                )

                latents = {
                    "gpt_cond_latent": gpt_cond_latent.cpu().squeeze().half().tolist(),
                    "speaker_embedding": speaker_embedding.cpu()
                    .squeeze()
                    .half()
                    .tolist(),
                }
                with open(filename_voice_json, "w") as new_file:
                    json.dump(latents, new_file)

                return gpt_cond_latent, speaker_embedding

            else:
                audio_path_list = []
                for filename in filenames:
                    # verify that filename ends with .wav
                    if filename.endswith(".wav"):
                        if voices_path:
                            filename_voice_wav = os.path.join(voices_path, filename)
                        else:
                            filename_voice_wav = filename
                        audio_path_list.append(filename_voice_wav)
                        logging.debug(
                            f"Added {filename_voice_wav} (#{len(audio_path_list)}) to audio_path_list"
                        )

                if len(audio_path_list) == 0:
                    logging.info(
                        "Using default female voice, no cloning source specified."
                    )

                    # Get the directory of the current script
                    current_dir = os.path.dirname(os.path.realpath(__file__))
                    filename_voice_json = os.path.join(
                        current_dir, "coqui_default_voice.json"
                    )
                    if not os.path.exists(filename_voice_json):
                        raise ValueError(
                            f"Default voice file {filename_voice_json} not found."
                        )

                # compute and write latents to json file
                logging.debug(f"Computing latents for {filename}")

                gpt_cond_latent, speaker_embedding = tts.get_conditioning_latents(
                    audio_path=audio_path_list, gpt_cond_len=30, max_ref_length=60
                )

                latents = {
                    "gpt_cond_latent": gpt_cond_latent.cpu().squeeze().half().tolist(),
                    "speaker_embedding": speaker_embedding.cpu()
                    .squeeze()
                    .half()
                    .tolist(),
                }
                filename_voice_json = audio_path_list[0][:-3] + "json"
                with open(filename_voice_json, "w") as new_file:
                    json.dump(latents, new_file)

                return gpt_cond_latent, speaker_embedding

        def postprocess_wave(chunk):
            """Post process the output waveform"""
            if isinstance(chunk, list):
                chunk = torch.cat(chunk, dim=0)
            chunk = chunk.clone().detach().cpu().numpy()
            chunk = chunk[None, : int(chunk.shape[0])]
            chunk = np.clip(chunk, -1, 1)
            chunk = chunk.astype(np.float32)
            return chunk

        def load_model(checkpoint, tts):
            global config
            try:
                if tts:
                    import gc

                    del tts
                    torch.cuda.empty_cache()
                    gc.collect()
                    from numba import cuda

                    current_device = cuda.get_current_device()
                    current_device.reset()
                    tts = None
                    import time

                    time.sleep(TIME_SLEEP_DEVICE_RESET)

                torch.set_num_threads(int(str(thread_count)))
                torch_device = torch.device(
                    "cuda"
                    if device == "cuda" and torch.cuda.is_available()
                    else "mps"
                    if device == "mps"
                    and torch.backends.mps.is_available()
                    and torch.backends.mps.is_built()
                    else "cuda"
                    if torch.cuda.is_available()
                    else "cpu"
                )

                config = load_config((os.path.join(checkpoint, "config.json")))
                tts = setup_tts_model(config)
                logging.debug(f"  xtts load_checkpoint({checkpoint})")

                print(f"LOADING CONFIG JSON FROM {os.path.join(checkpoint, 'config.json')}")

                tts.load_checkpoint(
                    config,
                    checkpoint_dir=checkpoint,
                    checkpoint_path=None,
                    vocab_path=None,
                    eval=True,
                    use_deepspeed=use_deepspeed,
                )
                tts.to(torch_device)
            except Exception as e:
                print(f"Error loading model for checkpoint {checkpoint}: {e}")
                raise
            return tts

        def get_user_data_dir(appname):
            TTS_HOME = os.environ.get("TTS_HOME")
            XDG_DATA_HOME = os.environ.get("XDG_DATA_HOME")
            if TTS_HOME is not None:
                ans = Path(TTS_HOME).expanduser().resolve(strict=False)
            elif XDG_DATA_HOME is not None:
                ans = Path(XDG_DATA_HOME).expanduser().resolve(strict=False)
            elif sys.platform == "win32":
                import winreg  # pylint: disable=import-outside-toplevel

                key = winreg.OpenKey(
                    winreg.HKEY_CURRENT_USER,
                    r"Software\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders",
                )
                dir_, _ = winreg.QueryValueEx(key, "Local AppData")
                ans = Path(dir_).resolve(strict=False)
            elif sys.platform == "darwin":
                ans = Path("~/Library/Application Support/").expanduser()
            else:
                ans = Path.home().joinpath(".local/share")
            return ans.joinpath(appname)

        logging.debug(f"Initializing coqui model {model_name}")
        logging.debug(f" - cloning reference {cloning_reference_wav}")
        logging.debug(f" - language {language}")
        logging.debug(f" - user data dir {get_user_data_dir('tts')}")
        logging.debug(f" - local model path {local_model_path}")

        try:
            path_tts = os.path.join(
                get_user_data_dir("tts"), model_name.replace("/", "--")
            )
            checkpoint = local_model_path if local_model_path else path_tts
            logging.debug(f" - checkpoint {checkpoint}")
            tts = load_model(checkpoint, tts)

            gpt_cond_latent, speaker_embedding = get_conditioning_latents(
                cloning_reference_wav, tts
            )

        except Exception as e:
            logging.exception(f"Error initializing main coqui engine model: {e}")
            raise

        ready_event.set()

        logging.info("Coqui text to speech synthesize model initialized successfully")

        try:
            while True:
                timeout = 0.1
                if conn.poll(timeout):  # Use poll with a tiny timeout to avoid blocking
                    try:
                        message = conn.recv()
                    except Exception as e:
                        logging.error(
                            f"conn.recv() error: {e} occurred in the "
                            "synthesize worker thread of Coqui engine."
                        )
                        time.sleep(1)
                        continue
                else:
                    # Poll timed out, continue without blocking
                    time.sleep(0.01)
                    continue

                command = message["command"]
                data = message["data"]

                if command == "update_reference":
                    new_wav_path = data["cloning_reference_wav"]
                    logging.info(f"Updating reference WAV to {new_wav_path}")
                    gpt_cond_latent, speaker_embedding = get_conditioning_latents(
                        new_wav_path, tts
                    )
                    conn.send(("success", "Reference updated successfully"))

                elif command == "set_speed":
                    speed = data["speed"]
                    conn.send(("success", "Speed updated successfully"))

                elif command == "set_model":
                    checkpoint = data["checkpoint"]
                    logging.info(f"Updating model checkpoint to {checkpoint}")
                    tts = load_model(checkpoint, tts)
                    conn.send(("success", "Model updated successfully"))

                elif command == "shutdown":
                    logging.info("Shutdown command received. Exiting worker process.")
                    conn.send(("shutdown", "shutdown"))
                    break  # This exits the loop, effectively stopping the worker process.

                elif command == "synthesize":
                    text = data["text"]
                    language = data["language"]

                    logging.debug(f"Starting inference for text: {text}")

                    time_start = time.time()
                    seconds_to_first_chunk = 0.0
                    full_generated_seconds = 0.0
                    raw_inference_start = 0.0
                    first_chunk_length_seconds = 0.0

                    chunks = tts.inference_stream(
                        text,
                        language,
                        gpt_cond_latent,
                        speaker_embedding,
                        stream_chunk_size=stream_chunk_size,
                        overlap_wav_len=overlap_wav_len,
                        temperature=temperature,
                        length_penalty=length_penalty,
                        repetition_penalty=repetition_penalty,
                        top_k=top_k,
                        top_p=top_p,
                        speed=speed,
                        enable_text_splitting=enable_text_splitting,
                    )

                    if full_sentences:
                        chunklist = []

                        for i, chunk in enumerate(chunks):
                            chunk = postprocess_wave(chunk)
                            chunk_bytes = chunk.tobytes()
                            chunklist.append(chunk_bytes)
                            chunk_duration = len(chunk_bytes) / (4 * 24000)
                            full_generated_seconds += chunk_duration
                            if i == 0:
                                first_chunk_length_seconds = chunk_duration
                                raw_inference_start = time.time()
                                seconds_to_first_chunk = (
                                    raw_inference_start - time_start
                                )

                        for i, chunk in enumerate(chunks):
                            chunk = postprocess_wave(chunk)
                            chunklist.append(chunk.tobytes())

                        for chunk in chunklist:
                            conn.send(("success", chunk))
                    else:
                        for i, chunk in enumerate(chunks):
                            chunk = postprocess_wave(chunk)
                            chunk_bytes = chunk.tobytes()
                            conn.send(("success", chunk_bytes))
                            chunk_duration = len(chunk_bytes) / (
                                4 * 24000
                            )  # 4 bytes per sample, 24000 Hz
                            full_generated_seconds += chunk_duration
                            if i == 0:
                                first_chunk_length_seconds = chunk_duration
                                raw_inference_start = time.time()
                                seconds_to_first_chunk = (
                                    raw_inference_start - time_start
                                )
                            else:
                                chunk_production_seconds = time.time() - time_start
                                generated_audio_seconds = full_generated_seconds

                                # wait only if we are faster than realtime, meaning
                                # that chunk_production_seconds is smaller than generated_audio_seconds
                                if load_balancing:
                                    if chunk_production_seconds < (generated_audio_seconds + load_balancing_buffer_length):
                                        waiting_time = generated_audio_seconds - chunk_production_seconds - load_balancing_cut_off
                                        if waiting_time > 0:
                                            print(f"Waiting for {waiting_time} seconds")
                                            time.sleep(waiting_time)

                    time_end = time.time()
                    seconds = time_end - time_start

                    if (
                        full_generated_seconds > 0
                        and (full_generated_seconds - first_chunk_length_seconds) > 0
                    ):
                        realtime_factor = seconds / full_generated_seconds
                        raw_inference_time = seconds - seconds_to_first_chunk
                        raw_inference_factor = raw_inference_time / (
                            full_generated_seconds - first_chunk_length_seconds
                        )
                        if print_realtime_factor:
                            print(f"Realtime Factor: {realtime_factor}")
                            print(f"Raw Inference Factor: {raw_inference_factor}")


                    # Send silent audio
                    sample_rate = config.audio.sample_rate

                    end_sentence_delimeters = ".!?…。¡¿"
                    mid_sentence_delimeters = ";:,\n()[]{}-“”„”—/|《》"

                    if text[-1] in end_sentence_delimeters:
                        silence_duration = sentence_silence_duration
                    elif text[-1] in mid_sentence_delimeters:
                        silence_duration = comma_silence_duration
                    else:
                        silence_duration = default_silence_duration

                    silent_samples = int(sample_rate * silence_duration)
                    silent_chunk = np.zeros(silent_samples, dtype=np.float32)
                    conn.send(("success", silent_chunk.tobytes()))

                    conn.send(("finished", ""))

                    end_time = time.time()

        except KeyboardInterrupt:
            logging.info("Keyboard interrupt received. " "Exiting worker process.")
            conn.send(("shutdown", "shutdown"))

        except Exception as e:
            logging.error(
                f"General synthesis error: {e} occured in "
                "synthesize worker thread of coqui engine."
            )

            tb_str = traceback.format_exc()
            print(f"Traceback: {tb_str}")
            print(f"Error: {e}")

            conn.send(("error", str(e)))
    
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

    def send_command(self, command, data):
        """
        Send a command to the worker process.
        """
        message = {"command": command, "data": data}
        self.parent_synthesize_pipe.send(message)

    def set_cloning_reference(self, cloning_reference_wav: str):
        """
        Send an 'update_reference' command and wait for a response.
        """
        if not isinstance(cloning_reference_wav, list):
            cloning_reference_wav = [cloning_reference_wav]
        self.send_command(
            "update_reference", {"cloning_reference_wav": cloning_reference_wav}
        )

        # Wait for the response from the worker process
        status, result = self.parent_synthesize_pipe.recv()
        if status == "success":
            logging.info("Reference WAV updated successfully")
        else:
            logging.error(f"Error updating reference WAV: {cloning_reference_wav}")

        return status, result

    def set_speed(self, speed: float):
        """
        Sets the speed of the speech synthesis.
        """
        self.send_command("set_speed", {"speed": speed})

        # Wait for the response from the worker process
        status, result = self.parent_synthesize_pipe.recv()
        if status == "success":
            logging.info("Speed updated successfully")
        else:
            logging.error("Error updating speed")

        return status, result

    def set_model(self, checkpoint: str):
        """
        Sets the model checkpoint
        """
        self.shutdown()

        self.specific_model = checkpoint
        self.model_path = self.download_model(checkpoint, self.local_models_path)

        self.create_worker_process()

    def get_stream_info(self):
        """
        Returns the PyAudio stream configuration
        information suitable for Coqui Engine.

        Returns:
            tuple: A tuple containing the audio format, number of channels,
              and the sample rate.
                  - Format (int): The format of the audio stream.
                    pyaudio.paInt16 represents 16-bit integers.
                  - Channels (int): The number of audio channels.
                    1 represents mono audio.
                  - Sample Rate (int): The sample rate of the audio in Hz.
                    16000 represents 16kHz sample rate.
        """
        return pyaudio.paFloat32, 1, 24000

    def _prepare_text_for_synthesis(self, text: str):
        """
        Splits a text into sentences.

        Args:
            text (str): Text to prepare for synthesis.

        Returns:
            text (str): Prepared text.
        """

        logging.debug(f'Preparing text for synthesis: "{text}"')

        if self.prepare_text_callback:
            return self.prepare_text_callback(text)

        # A fast fix for last character, may produce weird sounds if it is with text
        text = text.strip()
        text = text.replace("</s>", "")
        # text = re.sub("```.*```", "", text, flags=re.DOTALL)
        # text = re.sub("`.*`", "", text, flags=re.DOTALL)
        text = re.sub("\\(.*?\\)", "", text, flags=re.DOTALL)
        text = text.replace("```", "")
        text = text.replace("...", " ")
        text = text.replace("»", "")
        text = text.replace("«", "")
        text = re.sub(" +", " ", text)
        # text= re.sub("([^\x00-\x7F]|\w)(\.|\。|\?)",r"\1 \2\2",text)
        # text= re.sub("([^\x00-\x7F]|\w)(\.|\。|\?)",r"\1 \2",text)

        try:
            if len(text) > 2 and text[-1] in ["."]:
                text = text[:-1]
            elif len(text) > 2 and text[-1] in ["!", "?", ","]:
                text = text[:-1] + " " + text[-1]
            elif len(text) > 3 and text[-2] in ["."]:
                text = text[:-2]
            elif len(text) > 3 and text[-2] in ["!", "?", ","]:
                text = text[:-2] + " " + text[-2]
            
        except Exception as e:
            logging.warning(
                f'Error fixing sentence end punctuation: {e}, Text: "{text}"'
            )

        text = text.strip()

        logging.debug(f'Text after preparation: "{text}"')

        return text

    def synthesize(self, text: str) -> bool:
        """
        Synthesizes text to audio stream.

        Args:
            text (str): Text to synthesize.
        """

        with self._synthesize_lock:
            if self.add_sentence_filter:
                text = self._prepare_text_for_synthesis(text)

            if len(text) < 1:
                return

            data = {"text": text, "language": self.language}
            self.send_command("synthesize", data)

            status, result = self.parent_synthesize_pipe.recv()

            while "finished" not in status:
                if "shutdown" in status or "error" in status:
                    if "error" in status:
                        logging.error(f"Error synthesizing text: {text}")
                        logging.error(f"Error: {result}")
                    return False
                self.queue.put(result)
                status, result = self.parent_synthesize_pipe.recv()

            return True

    @staticmethod
    def download_file(url, destination):
        response = requests.get(url, stream=True)
        total_size_in_bytes = int(response.headers.get("content-length", 0))
        block_size = 1024

        progress_bar = tqdm(total=total_size_in_bytes, unit="iB", unit_scale=True)

        with open(destination, "wb") as file:
            for data in response.iter_content(block_size):
                progress_bar.update(len(data))
                file.write(data)

        progress_bar.close()

    @staticmethod
    def download_model(model_name="v2.0.2", local_models_path=None):
        # Creating a unique folder for each model version
        if local_models_path and len(local_models_path) > 0:
            model_folder = os.path.join(local_models_path, f"{model_name}")
            logging.info(f'Local models path: "{model_folder}"')
        else:
            model_folder = os.path.join(os.getcwd(), "models", f"{model_name}")
            logging.info(
                f'Checking for models within application directory: "{model_folder}"'
            )

        os.makedirs(model_folder, exist_ok=True)

        files = {
            "config.json": f"https://huggingface.co/coqui/XTTS-v2/raw/{model_name}/config.json",
            "model.pth": f"https://huggingface.co/coqui/XTTS-v2/resolve/{model_name}/model.pth?download=true",
            "vocab.json": f"https://huggingface.co/coqui/XTTS-v2/raw/{model_name}/vocab.json",
            "speakers_xtts.pth": f"https://huggingface.co/coqui/XTTS-v2/resolve/{model_name}/speakers_xtts.pth",
        }

        for file_name, url in files.items():
            file_path = os.path.join(model_folder, file_name)
            if not os.path.exists(file_path):
                print(f"Downloading {file_name} to {file_path}...")
                CoquiEngine.download_file(url, file_path)
                logging.info(f"{file_name} downloaded successfully.")
            else:
                logging.info(f"{file_name} exists in {file_path} (no download).")

        return model_folder

    def get_voices(self):
        """
        Retrieves the installed voices available for the Coqui TTS engine.
        """

        voice_objects = []
        voices_appended = []

        # Add custom voices
        files = os.listdir(self.voices_path)
        for file in files:
            # remove ending .wav or .json from filename
            if file.endswith(".wav"):
                file = file[:-4]
            elif file.endswith(".json"):
                file = file[:-5]
            else:
                continue

            if file in voices_appended:
                continue

            voices_appended.append(file)
            voice_objects.append(CoquiVoice(file))

        # Add predefined coqui system voices
        for voice in self.voices_list:
            voice_objects.append(CoquiVoice(voice))

        return voice_objects

    def set_voice(self, voice: str):
        """
        Sets the voice to be used for speech synthesis.
        """
        if isinstance(voice, CoquiVoice):
            self.set_cloning_reference(voice.name)
        else:
            installed_voices = self.get_voices()
            for installed_voice in installed_voices:
                if voice in installed_voice.name:
                    self.set_cloning_reference(installed_voice.name)
                    return
            self.set_cloning_reference(voice)

    def set_voice_parameters(self, **voice_parameters):
        """
        Sets the voice parameters to be used for speech synthesis.

        Args:
            **voice_parameters: The voice parameters to be used for speech synthesis.

        This method should be overridden by the derived class to set the desired voice parameters.
        """
        pass

    def shutdown(self):
        """
        Shuts down the engine by terminating the process and closing the pipes.
        """
        # Send shutdown command to the worker process
        logging.info("Sending shutdown command to the worker process")
        self.send_command("shutdown", {})

        self.output_queue.put("STOP")
        self.output_worker_thread.join()

        # Wait for the worker process to acknowledge the shutdown
        try:
            status, _ = self.parent_synthesize_pipe.recv()
            if "shutdown" in status:
                logging.info("Worker process acknowledged shutdown")
        except EOFError:
            # Pipe was closed, meaning the process is already down
            logging.warning(
                "Worker process pipe was closed before shutdown acknowledgement"
            )

        # Close the pipe connection
        self.parent_synthesize_pipe.close()

        # Terminate the process
        logging.info("Terminating the worker process")
        self.synthesize_process.terminate()

        # Wait for the process to terminate
        self.synthesize_process.join()
        logging.info("Worker process has been terminated")

    def retrieve_coqui_voices(self):
        """
        Retrieves the installed voices available for the Coqui TTS engine.
        """

        if self.pretrained:
            return self.voices_list

        from TTS.tts.layers.xtts.xtts_manager import SpeakerManager

        if self.local_models_path:
            local_models_path = self.local_models_path
        else:
            local_models_path = os.path.join(
                os.getcwd(), "models", f"{self.specific_model}"
            )

        speaker_file_path = os.path.join(local_models_path, "speakers_xtts.pth")
        if not os.path.exists(speaker_file_path):
            speaker_file_path = os.path.join(
                local_models_path, self.specific_model, "speakers_xtts.pth"
            )

        speaker_manager = SpeakerManager(speaker_file_path)
        self.voices_list = []
        for speaker_name in speaker_manager.name_to_id:
            self.voices_list.append(speaker_name)
        return self.voices_list



import queue
import threading
import uuid
from typing import Generator, List, Any

class BufferStream:
    def __init__(self):
        self.items: queue.Queue = queue.Queue()
        self._stop_event: threading.Event = threading.Event()
        self.stopped: bool = False
        self.stream_id: str = str(uuid.uuid4())

    def add(self, item: Any) -> None:
        """Add an item to the buffer."""
        self.items.put(item)

    def stop(self) -> None:
        """Signal to stop the buffer stream."""
        self._stop_event.set()

    def snapshot(self) -> List[Any]:
        """Take a snapshot of all items in the buffer without exhausting it."""
        with self.items.mutex:
            return list(self.items.queue)

    def gen(self) -> Generator[Any, None, None]:
        """Generate items from the buffer, yielding them one at a time."""
        while not self._stop_event.is_set() or not self.items.empty():
            try:
                yield self.items.get(timeout=0.1)
            except queue.Empty:
                continue
        self.stopped = True



from parler_tts import ParlerTTSForConditionalGeneration, ParlerTTSStreamer
from transformers import AutoTokenizer
from .base_engine import BaseEngine
from threading import Thread
from typing import Union
import pyaudio
import torch


class ParlerVoice:
    def __init__(self, name, description):
        self.name = name
        self.description = description

    def __repr__(self):
        return self.name

class ParlerEngine(BaseEngine):
    def __init__(
        self,
        voice_prompt="Laura's voice is bright, joyful, and full of energy. Each word is engaging, radiating warmth and positivity that lifts the listener's mood.",
        model_name="parler-tts/parler-tts-mini-v1", # ylacombe/parler-tts-mini-jenny-30H
        device="cuda:0",
        torch_dtype=torch.bfloat16,
        buffer_duration_s=1.0,
        play_steps_in_s=0.5,
    ):
        """
        Initializes the Parler TTS engine.

        Args:
            model_name (str): Name of the Parler TTS model to use.
            device (str): Torch device to use (e.g., "cuda:0", "cpu").
            torch_dtype (torch.dtype): Torch data type to use.
            voice_prompt (str): Voice prompt for the model.
            play_steps_in_s (float): Duration in seconds for each play step.
        """
        super().__init__()
        self.model_name = model_name
        self.device = device
        self.torch_dtype = torch_dtype
        self.voice_prompt = voice_prompt
        self.play_steps_in_s = play_steps_in_s
        self.voice_parameters = {}
        self.buffer_duration_s = buffer_duration_s

        self.initialize_model()

    def post_init(self):
        self.engine_name = "parler_tts"

    def initialize_model(self):
        """
        Loads the model and tokenizer.
        """
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_name)
        self.model = ParlerTTSForConditionalGeneration.from_pretrained(
            self.model_name
        ).to(self.device, dtype=self.torch_dtype)

    def get_stream_info(self):
        """
        Returns the PyAudio stream configuration information suitable for Parler TTS Engine.

        Returns:
            tuple: A tuple containing the audio format, number of channels, and the sample rate.
        """
        return pyaudio.paFloat32, 1, 44100  # Using 32-bit float, mono, 44100 Hz

    def synthesize(self, text: str) -> bool:
        """
        Synthesizes text to audio stream.

        Args:
            text (str): Text to synthesize.

        Returns:
            bool: True if synthesis starts successfully, False otherwise.
        """
        try:
            self._generate_and_queue_audio(text)
            return True
        except Exception as e:
            print(f"Error in synthesize: {e}")
            return False

    def _generate_and_queue_audio(self, text: str):
        """
        Generates audio from text, buffers audio for a set duration, and then streams it into the audio queue.

        Args:
            text (str): Text to synthesize.
        """
        frame_rate = self.model.audio_encoder.config.frame_rate
        sampling_rate = self.model.audio_encoder.config.sampling_rate

        play_steps = int(frame_rate * self.play_steps_in_s)
        streamer = ParlerTTSStreamer(self.model, device=self.device, play_steps=play_steps)

        if not self.voice_prompt:
            self.voice_prompt = "Female voice, calm and uplifting."

        inputs = self.tokenizer(self.voice_prompt, return_tensors="pt").to(self.device)
        prompt = self.tokenizer(text, return_tensors="pt").to(self.device)

        generation_kwargs = {
            "input_ids": inputs.input_ids,
            "prompt_input_ids": prompt.input_ids,
            "attention_mask": inputs.attention_mask,
            "prompt_attention_mask": prompt.attention_mask,
            "streamer": streamer,
            "do_sample": True,
            "temperature": 1.0,
            "min_new_tokens": 10,
            **self.voice_parameters,  # Merge with any additional voice parameters
        }

        # Initialize variables for buffering
        audio_buffer = []
        buffer_length_s = 0.0
        generation_completed = False

        # Start the audio generation (blocking call)
        def generate_audio():
            self.model.generate(**generation_kwargs)

        # Start the generation in a separate thread
        generation_thread = Thread(target=generate_audio)
        generation_thread.start()

        # Process the streamer in the main thread
        while not generation_completed:
            try:
                new_audio = next(streamer)
                if new_audio.shape[0] == 0:
                    # Streamer signaled completion
                    generation_completed = True
                    break

                audio_chunk = new_audio
                audio_buffer.append(audio_chunk)
                buffer_length_s += new_audio.shape[0] / sampling_rate

                if buffer_length_s >= self.buffer_duration_s:
                    # Buffering complete, start streaming
                    break
            except StopIteration:
                # No more audio data
                generation_completed = True
                break

        # Queue the buffered audio chunks
        for buffered_chunk in audio_buffer:
            self.queue.put(buffered_chunk.tobytes())

        # Continue streaming the rest of the audio
        while not generation_completed:
            try:
                new_audio = next(streamer)
                if new_audio.shape[0] == 0:
                    # Streamer signaled completion
                    generation_completed = True
                    break
                audio_chunk = new_audio
                self.queue.put(audio_chunk.tobytes())
            except StopIteration:
                generation_completed = True
                break

        # Ensure the generation thread has completed
        generation_thread.join()

    def get_voices(self):
        """
        Retrieves the voices available from Parler TTS.

        Returns:
            list: A list containing ParlerVoice objects representing each available voice.
        """
        return [
            ParlerVoice(
                "Laura",
                "Laura's voice is fast-paced with a slight rasp, recorded up close with no background noise. Her words come out quickly, creating a sense of urgency and intensity.",
            ),
            ParlerVoice(
                "John",
                "John's voice is deep and smooth, with a calm and reassuring tone. He speaks slowly and clearly, with a slight southern accent.",
            ),
        ]

    def set_voice(self, voice: Union[str, ParlerVoice]):
        """
        Sets the voice to be used for speech synthesis.

        Args:
            voice (Union[str, ParlerVoice]): The voice to be used for speech synthesis.
        """
        if isinstance(voice, ParlerVoice):
            self.voice_prompt = voice.description
        elif isinstance(voice, str):
            self.voice_prompt = voice
        else:
            raise ValueError("Voice must be a string or ParlerVoice instance.")

    def set_voice_parameters(self, **voice_parameters):
        """
        Sets additional voice parameters for speech synthesis.

        Args:
            **voice_parameters: Arbitrary keyword arguments for voice parameters.
        """
        self.voice_parameters.update(voice_parameters)

    def shutdown(self):
        """
        Shuts down the engine.
        """
        pass  



import os
import sys
import subprocess

def main():
    # Get the current script's directory (should be tts-cli)
    script_dir = os.path.dirname(os.path.abspath(__file__))
    
    # Move one directory up to access the venv
    root_dir = os.path.dirname(script_dir)
    os.chdir(root_dir)
    
    # Path to the virtual environment
    venv_path = os.path.join(root_dir, 'venv')
    
    # Path to the Python interpreter in the virtual environment
    if sys.platform == "win32":
        python_path = os.path.join(venv_path, 'Scripts', 'python.exe')
    else:
        python_path = os.path.join(venv_path, 'bin', 'python')
    
    # Change back to the tts-cli directory
    os.chdir(script_dir)
    
    # Prepare the command to run tts_server.py with all provided arguments
    command = [python_path, 'tts_server.py'] + sys.argv[1:]
    
    # Start the TTS server
    print("Starting TTS server...")
    print(f"Command: {command}")
    try:
        subprocess.run(command, check=True)
    except subprocess.CalledProcessError as e:
        print(f"Error starting TTS server: {e}")
        sys.exit(1)
    except FileNotFoundError:
        print(f"Error: Could not find Python interpreter at {python_path}")
        print("Make sure the virtual environment is set up correctly.")
        sys.exit(1)

if __name__ == "__main__":
    main()

# import os
# import sys
# import subprocess

# def start_tts_server():
#     # Get the current script's directory (should be tts-cli)
#     script_dir = os.path.dirname(os.path.abspath(__file__))
    
#     # Move one directory up to access the venv
#     root_dir = os.path.dirname(script_dir)
#     os.chdir(root_dir)
    
#     # Path to the virtual environment
#     venv_path = os.path.join(root_dir, 'venv')
    
#     # Path to the Python interpreter in the virtual environment
#     if sys.platform == "win32":
#         python_path = os.path.join(venv_path, 'Scripts', 'python.exe')
#     else:
#         python_path = os.path.join(venv_path, 'bin', 'python')
    
#     # Change back to the tts-cli directory
#     os.chdir(script_dir)
    
#     # Start the TTS server
#     print("Starting TTS server...")
#     try:
#         subprocess.run([python_path, 'tts_server.py'], check=True)
#     except subprocess.CalledProcessError as e:
#         print(f"Error starting TTS server: {e}")
#     except FileNotFoundError:
#         print(f"Error: Could not find Python interpreter at {python_path}")
#         print("Make sure the virtual environment is set up correctly.")

# def main():
#     start_tts_server()

# if __name__ == "__main__":
#     main()



from .base_engine import BaseEngine  # noqa: F401

# Optional dependencies
try:
    from .azure_engine import AzureEngine, AzureVoice  # noqa: F401
    import azure.cognitiveservices.speech as tts
except ImportError:
    AzureEngine, AzureVoice, tts = None, None, None

try:
    from .system_engine import SystemEngine, SystemVoice  # noqa: F401
    import pyttsx3
except ImportError:
    SystemEngine, SystemVoice, pyttsx3 = None, None, None

try:
    from .elevenlabs_engine import ElevenlabsEngine, ElevenlabsVoice  # noqa: F401
    from elevenlabs.client import ElevenLabs
except ImportError:
    ElevenlabsEngine, ElevenlabsVoice, ElevenLabs = None, None, None

try:
    from .coqui_engine import CoquiEngine, CoquiVoice  # noqa: F401
    from TTS.utils.manage import ModelManager
except ImportError:
    CoquiEngine, CoquiVoice, ModelManager = None, None, None

try:
    from .openai_engine import OpenAIEngine, OpenAIVoice  # noqa: F401
    from openai import OpenAI
except ImportError:
    OpenAIEngine, OpenAIVoice, OpenAI = None, None, None

try:
    from .gtts_engine import GTTSEngine, GTTSVoice  # noqa: F401
    from gtts import gTTS
except ImportError:
    GTTSEngine, GTTSVoice, gTTS = None, None, None

try:
    from .parler_engine import ParlerEngine, ParlerVoice  # noqa: F401
except ImportError as e:
    ParlerEngine, ParlerVoice = None, None



from .base_engine import BaseEngine
from pydub import AudioSegment
from typing import Union
import tempfile
import wave
import os
import io
import pyaudio
from gtts import gTTS
import gtts.lang

SYNTHESIS_FILE = "gtts_speech_synthesis.wav"


class GTTSVoice:
    def __init__(
        self,
        language: str = "en",
        tld: str = "com",
        chunk_length: int = 100,
        crossfade_length: int = 10,
        speed_increase: float = 1.0,
    ):
        self.language = language
        self.tld = tld
        self.chunk_length = chunk_length
        self.crossfade_length = crossfade_length
        self.speed_increase = speed_increase

    def __repr__(self):
        return f"{self.language} ({self.tld}), Chunk: {self.chunk_length}, Crossfade: {self.crossfade_length}, Speedup: {self.speed_increase}"


class GTTSEngine(BaseEngine):
    def __init__(
        self,
        voice: Union[str, GTTSVoice] = GTTSVoice("en", "com"),
        print_installed_voices: bool = False,
    ):
        """
        Initializes a gTTS text-to-speech engine object.

        Args:
            voice (Union[str, GTTSVoice], optional): Voice configuration. Defaults to GTTSVoice("en", "com").
            print_installed_voices (bool, optional): Indicates if the list of available languages should be printed. Defaults to False.
        """
        self.set_voice(voice)
        temp_file_path = tempfile.gettempdir()
        self.file_path = os.path.join(temp_file_path, SYNTHESIS_FILE)

        if print_installed_voices:
            print(self.get_voices())

    def post_init(self):
        self.engine_name = "gtts"

    def get_stream_info(self):
        """
        Returns the PyAudio stream configuration information suitable for gTTS Engine.

        Returns:
            tuple: A tuple containing the audio format, number of channels, and the sample rate.
                  - Format (int): The format of the audio stream. pyaudio.paInt16 represents 16-bit integers.
                  - Channels (int): The number of audio channels. 1 represents mono audio.
                  - Sample Rate (int): The sample rate of the audio in Hz. 22050 represents 22.05kHz sample rate.
        """
        return pyaudio.paInt16, 1, 22050

    def synthesize(self, text: str) -> bool:
        """
        Synthesizes text to audio stream.

        Args:
            text (str): Text to synthesize.
        """
        try:
            # Generate audio with gTTS
            with io.BytesIO() as f:
                tts = gTTS(text=text, lang=self.voice.language, tld=self.voice.tld)
                tts.write_to_fp(f)
                f.seek(0)

                audio = AudioSegment.from_file(f, format="mp3")

                if self.voice.speed_increase != 1.0:
                    audio = audio.speedup(
                        playback_speed=self.voice.speed_increase,
                        chunk_size=self.voice.chunk_length,
                        crossfade=self.voice.crossfade_length,
                    )

                audio.export(self.file_path, format="wav")

            # Now open the WAV file and read the chunks
            with wave.open(self.file_path, "rb") as wf:
                audio_data = wf.readframes(wf.getnframes())
                self.queue.put(audio_data)
                return True

        except Exception as e:
            print(f"Error in synthesizing text: {e}")
            return False

    def get_voices(self):
        """
        Retrieves the voices available from gTTS.

        Returns:
            list[GTTSVoice]: A list containing GTTSVoice objects representing each available language and TLD.
        """
        voices = []
        languages = gtts.lang.tts_langs()
        tlds = ["com", "com.au", "co.uk", "us", "ca", "co.in", "ie", "co.za"]

        for lang in languages.keys():
            for tld in tlds:
                voices.append(GTTSVoice(language=lang, tld=tld))

        return voices

    def set_voice(self, voice: Union[str, GTTSVoice]):
        """
        Sets the voice to be used for speech synthesis.

        Args:
            voice (Union[str, GTTSVoice]): The voice to be used for speech synthesis.
        """
        if isinstance(voice, GTTSVoice):
            self.voice = voice
        else:
            self.voice = GTTSVoice(language=voice, tld="com")



# threadsafe_generators.py

"""
This file contains a collection of classes aimed at providing thread-safe operations over generators
and iterables. The utility of this module can be mainly seen in multi-threaded environments where generators 
or iterables need to be consumed across threads without race conditions. Additionally, functionalities like 
character-based iteration and accumulation are provided for enhanced flexibility.
"""

from typing import Union, Iterator, Callable, Optional
import threading
from dataclasses import dataclass, field


@dataclass
class CharIterator:
    """
    An iterator that allows iteration over characters of strings or string iterators.
    
    Attributes:
        items (List[Union[str, Iterator[str]]]): The list of strings or string iterators.
        _index (int): Current index in the items list being iterated.
        _char_index (Optional[int]): Current character index in the current string.
        _current_iterator (Optional[Iterator[str]]): Current iterator being consumed.
        immediate_stop (threading.Event): Event signaling to stop iteration.
        iterated_text (str): Accumulates the characters that have been iterated over.
        log_characters (bool): If True, logs processed characters.
        on_character (Callable): Callback on each character processed.
        on_first_text_chunk (Callable): Callback on receiving the first text chunk.
        on_last_text_chunk (Callable): Callback on receiving the last text chunk.
        first_chunk_received (bool): Flag indicating if the first chunk was processed.
    """

    log_characters: bool = False
    on_character: Optional[Callable[[str], None]] = None
    on_first_text_chunk: Optional[Callable[[], None]] = None
    on_last_text_chunk: Optional[Callable[[], None]] = None

    items: list = field(default_factory=list)
    _index: int = 0
    _char_index: Optional[int] = None
    _current_iterator: Optional[Iterator[str]] = None
    immediate_stop: threading.Event = field(default_factory=threading.Event)
    iterated_text: str = ""
    first_chunk_received: bool = False

    def add(self, item: Union[str, Iterator[str]]) -> None:
        """Add a string or a string iterator to the list of items."""
        self.items.append(item)

    def stop(self) -> None:
        """Signal the iterator to stop immediately during the next iteration."""
        self.immediate_stop.set()

    def __iter__(self) -> "CharIterator":
        """Return the iterator object itself."""
        return self

    def _log_and_trigger(self, char: str) -> None:
        """Log character and trigger associated callbacks."""
        self.iterated_text += char
        if self.log_characters:
            print(char, end="", flush=True)
        if self.on_character:
            self.on_character(char)
        if not self.first_chunk_received and self.on_first_text_chunk:
            self.on_first_text_chunk()
            self.first_chunk_received = True

    def __next__(self) -> str:
        """Fetch the next character from the current string or string iterator."""
        if self.immediate_stop.is_set():
            raise StopIteration

        while self._index < len(self.items):
            item = self.items[self._index]

            if isinstance(item, str):
                if self._char_index is None:
                    self._char_index = 0

                if self._char_index < len(item):
                    char = item[self._char_index]
                    self._char_index += 1
                    self._log_and_trigger(char)
                    return char
                else:
                    self._char_index = None
                    self._index += 1

            else:  # item is an iterator
                if self._current_iterator is None:
                    self._current_iterator = iter(item)

                if self._char_index is None:
                    try:
                        self._current_str = next(self._current_iterator)
                        if hasattr(self._current_str, "choices"):
                            self._current_str = str(self._current_str.choices[0].delta.content) or ""
                    except StopIteration:
                        self._char_index = None
                        self._current_iterator = None
                        self._index += 1
                        continue

                    self._char_index = 0

                if self._char_index < len(self._current_str):
                    char = self._current_str[self._char_index]
                    self._char_index += 1
                    self._log_and_trigger(char)
                    return char
                else:
                    self._char_index = None

        if self.iterated_text and self.on_last_text_chunk:
            self.on_last_text_chunk()

        raise StopIteration


class AccumulatingThreadSafeGenerator:
    """
    A thread-safe generator that accumulates the iterated tokens into a text.
    """

    def __init__(self, gen_func: Iterator[str], on_first_text_chunk: Optional[Callable[[], None]] = None, on_last_text_chunk: Optional[Callable[[], None]] = None):
        """
        Initialize the AccumulatingThreadSafeGenerator instance.
        
        Args:
            gen_func (Iterator[str]): The generator function to be used.
            on_first_text_chunk (Optional[Callable]): Callback for the first chunk of text.
            on_last_text_chunk (Optional[Callable]): Callback for the last chunk of text.
        """
        self.lock = threading.Lock()
        self.generator = gen_func
        self.exhausted = False
        self.iterated_text = ""
        self.on_first_text_chunk = on_first_text_chunk
        self.on_last_text_chunk = on_last_text_chunk
        self.first_chunk_received = False

    def __iter__(self) -> "AccumulatingThreadSafeGenerator":
        """Return the iterator object itself."""
        return self

    def __next__(self) -> str:
        """Fetch the next token from the generator in a thread-safe manner."""
        with self.lock:
            try:
                token = next(self.generator)
                self.iterated_text += str(token)

                if not self.first_chunk_received and self.on_first_text_chunk:
                    self.on_first_text_chunk()

                self.first_chunk_received = True
                return token

            except StopIteration:
                if self.iterated_text and self.on_last_text_chunk:
                    self.on_last_text_chunk()
                self.exhausted = True
                raise

    def is_exhausted(self) -> bool:
        """Check if the generator has been exhausted."""
        with self.lock:
            return self.exhausted

    def accumulated_text(self) -> str:
        """Retrieve the accumulated text from the iterated tokens."""
        with self.lock:
            return self.iterated_text



from abc import ABCMeta, ABC
from typing import Union
import shutil
import queue


# Define a meta class that will automatically call the BaseEngine's __init__ method
# and also the post_init method if it exists.
class BaseInitMeta(ABCMeta):
    def __call__(cls, *args, **kwargs):
        # Create an instance of the class that this meta class is used on.
        instance = super().__call__(*args, **kwargs)

        # Call the __init__ method of BaseEngine to set default properties.
        BaseEngine.__init__(instance)

        # If the instance has a post_init method, call it.
        # This allows subclasses to define additional initialization steps.
        if hasattr(instance, "post_init"):
            instance.post_init()

        return instance


# Define a base class for engines with the custom meta class.
class BaseEngine(ABC, metaclass=BaseInitMeta):
    def __init__(self):
        self.engine_name = "unknown"

        # Indicates if the engine can handle generators.
        self.can_consume_generators = False

        # Queue to manage tasks or data for the engine.
        self.queue = queue.Queue()

        # Callback to be called when an audio chunk is available.
        self.on_audio_chunk = None

        # Callback to be called when the engine is starting to synthesize audio.
        self.on_playback_start = None

    def get_stream_info(self):
        """
        Returns the audio stream configuration information suitable for PyAudio.

        Returns:
            tuple: A tuple containing the audio format, number of channels, and the sample rate.
                  - Format (int): The format of the audio stream. pyaudio.paInt16 represents 16-bit integers.
                  - Channels (int): The number of audio channels. 1 represents mono audio.
                  - Sample Rate (int): The sample rate of the audio in Hz. 16000 represents 16kHz sample rate.
        """
        raise NotImplementedError(
            "The get_stream_info method must be implemented by the derived class."
        )

    def synthesize(self, text: str) -> bool:
        """
        Synthesizes text to audio stream.

        Args:
            text (str): Text to synthesize.
        """
        raise NotImplementedError(
            "The synthesize method must be implemented by the derived class."
        )

    def get_voices(self):
        """
        Retrieves the voices available from the specific voice source.

        This method should be overridden by the derived class to fetch the list of available voices.

        Returns:
            list: A list containing voice objects representing each available voice.
        """
        raise NotImplementedError(
            "The get_voices method must be implemented by the derived class."
        )

    def set_voice(self, voice: Union[str, object]):
        """
        Sets the voice to be used for speech synthesis.

        Args:
            voice (Union[str, object]): The voice to be used for speech synthesis.

        This method should be overridden by the derived class to set the desired voice.
        """
        raise NotImplementedError(
            "The set_voice method must be implemented by the derived class."
        )

    def set_voice_parameters(self, **voice_parameters):
        """
        Sets the voice parameters to be used for speech synthesis.

        Args:
            **voice_parameters: The voice parameters to be used for speech synthesis.

        This method should be overridden by the derived class to set the desired voice parameters.
        """
        raise NotImplementedError(
            "The set_voice_parameters method must be implemented by the derived class."
        )

    def shutdown(self):
        """
        Shuts down the engine.
        """
        pass

    def is_installed(self, lib_name: str) -> bool:
        """
        Check if the given library or software is installed and accessible.

        This method uses shutil.which to determine if the given library or software is
        installed and available in the system's PATH.

        Args:
            lib_name (str): Name of the library or software to check.

        Returns:
            bool: True if the library is installed, otherwise False.
        """
        lib = shutil.which(lib_name)
        if lib is None:
            return False
        return True



"""
Python script to download the ai models
needed by linguflex from Huggingface's model hub.
"""

from huggingface_hub import hf_hub_download
import os


def create_directory(path):
    if not os.path.exists(path):
        os.makedirs(path)


def create_directories():
    create_directory("assets")
    create_directory("assets/hubert")
    create_directory("models")
    create_directory("models/rvc")
    create_directory("models/xtts")
    create_directory("models/xtts/Lasinya")


def download_file(
        url,
        filename_server,
        path_local
        ):

    local_file = os.path.join(path_local, filename_server)
    if os.path.exists(local_file):
        print(f"File {filename_server} already exists in {path_local}.")
        return

    print(f"Downloading {filename_server} from repo {url} to {path_local}")
    hf_hub_download(
        repo_id=url,
        filename=filename_server,
        local_dir=path_local)


create_directories()

# download rvc base model (hubert) files
print("Downloading hubert base model files")
download_file(
    "KoljaB/RVC_Assets", "hubert_base.pt", "assets/hubert")
download_file(
    "KoljaB/RVC_Assets", "hubert_inputs.pth", "assets/hubert")

# download rvc trained model files
print("Downloading rvc trained model files")
download_file(
    "KoljaB/RVC_Models", "Lasinya.pth", "models/rvc")
download_file(
    "KoljaB/RVC_Models", "Lasinya.index", "models/rvc")

# download xtts trained model files
from huggingface_hub import hf_hub_download
import os

def create_directory(path):
    if not os.path.exists(path):
        os.makedirs(path)

def create_directories():
    create_directory("assets")
    create_directory("assets/hubert")
    create_directory("models")
    create_directory("models/rvc")
    create_directory("models/xtts")
    create_directory("models/xtts/v2.0.2")

def download_file(url, filename_server, path_local):
    local_file = os.path.join(path_local, filename_server)
    if os.path.exists(local_file):
        print(f"File {filename_server} already exists in {path_local}.")
        return

    print(f"Downloading {filename_server} from repo {url} to {path_local}")
    hf_hub_download(
        repo_id=url,
        filename=filename_server,
        local_dir=path_local)

create_directories()

# download rvc base model (hubert) files
print("Downloading hubert base model files")
download_file(
    "KoljaB/RVC_Assets", "hubert_base.pt", "assets/hubert")
download_file(
    "KoljaB/RVC_Assets", "hubert_inputs.pth", "assets/hubert")

# download rvc trained model files
print("Downloading rvc trained model files")
download_file(
    "KoljaB/RVC_Models", "Lasinya.pth", "models/rvc")
download_file(
    "KoljaB/RVC_Models", "Lasinya.index", "models/rvc")

# download xtts v2 base model files
print("Downloading XTTS v2 base model files")
download_file(
     "coqui/XTTS-v2", "config.json", "models/xtts/v2.0.2")
download_file(
     "coqui/XTTS-v2", "model.pth", "models/xtts/v2.0.2")
download_file(
     "coqui/XTTS-v2", "vocab.json", "models/xtts/v2.0.2")
download_file(
     "coqui/XTTS-v2", "speakers_xtts.pth", "models/xtts/v2.0.2")
     

# print("Downloading xtts trained model files (Lasinya)")
# download_file(
#      "KoljaB/XTTS_Lasinya", "config.json", "models/xtts/Lasinya")
# download_file(
#      "KoljaB/XTTS_Lasinya", "vocab.json", "models/xtts/Lasinya")
# download_file(
#      "KoljaB/XTTS_Lasinya", "speakers_xtts.pth", "models/xtts/Lasinya")
# download_file(
#      "KoljaB/XTTS_Lasinya", "model.pth", "models/xtts/Lasinya")



from .threadsafe_generators import CharIterator, AccumulatingThreadSafeGenerator
from .stream_player import StreamPlayer, AudioConfiguration
from typing import Union, Iterator, List
from .engines import BaseEngine
import stream2sentence as s2s
import numpy as np
import threading
import traceback
import logging
import pyaudio
import queue
import time
import wave


class TextToAudioStream:
    def __init__(
        self,
        engine: Union[BaseEngine, List[BaseEngine]],
        log_characters: bool = False,
        on_text_stream_start=None,
        on_text_stream_stop=None,
        on_audio_stream_start=None,
        on_audio_stream_stop=None,
        on_character=None,
        output_device_index=None,
        tokenizer: str = "nltk",
        language: str = "en",
        muted: bool = False,
        level=logging.WARNING,
    ):
        """
        Initializes the TextToAudioStream.

        Args:
            engine (BaseEngine): The engine used for text to audio synthesis.
            log_characters (bool, optional): If True, logs the characters
              processed for synthesis.
            on_text_stream_start (callable, optional): Callback function that
              gets called when the text stream starts.
            on_text_stream_stop (callable, optional): Callback function that
              gets called when the text stream stops.
            on_audio_stream_start (callable, optional): Callback function that
              gets called when the audio stream starts.
            on_audio_stream_stop (callable, optional): Callback function that
              gets called when the audio stream stops.
            on_character (callable, optional): Callback function that gets
              called when a single character is processed.
            output_device_index (int, optional): The index of the output device
                to use for audio playback.
            tokenizer (str, optional): Tokenizer to use for sentence splitting
                (currently "nltk" and "stanza" are supported).
            language (str, optional): Language to use for sentence splitting.
            muted (bool, optional): If True, disables audio playback via local
                speakers (in case you want to synthesize to file or process
                audio chunks). Default is False.
                If set to True it will override the play parameters muted
                setting.
            level (int, optional): Logging level. Defaults to logging.WARNING.
        """
        self.log_characters = log_characters
        self.on_text_stream_start = on_text_stream_start
        self.on_text_stream_stop = on_text_stream_stop
        self.on_audio_stream_start = on_audio_stream_start
        self.on_audio_stream_stop = on_audio_stream_stop
        self.output_device_index = output_device_index
        self.output_wavfile = None
        self.chunk_callback = None
        self.wf = None
        self.abort_events = []
        self.tokenizer = tokenizer
        self.language = language
        self.global_muted = muted
        self.player = None
        self.play_lock = threading.Lock()
        self.is_playing_flag = False

        self._create_iterators()

        logging.info(f"Initializing tokenizer {tokenizer} " f"for language {language}")
        s2s.init_tokenizer(tokenizer, language)

        # Initialize the play_thread attribute
        # (used for playing audio in a separate thread)
        self.play_thread = None

        # Initialize an attribute to store generated text
        self.generated_text = ""

        # A flag to indicate if the audio stream is currently running or not
        self.stream_running = False

        self.on_character = on_character

        self.engine_index = 0
        if isinstance(engine, list):
            # Handle the case where engine is a list of BaseEngine instances
            self.engines = engine
        else:
            # Handle the case where engine is a single BaseEngine instance
            self.engines = [engine]

        self.load_engine(self.engines[self.engine_index])

    def load_engine(self, engine: BaseEngine):
        """
        Loads the synthesis engine and prepares the audio player for stream playback.
        This method sets up the engine that will be used for text-to-audio conversion, extracts the necessary stream information like format, channels, and rate from the engine, and initializes the StreamPlayer if the engine does not support consuming generators directly.

        Args:
            engine (BaseEngine): The synthesis engine to be used for converting text to audio.
        """

        # Store the engine instance (responsible for text-to-audio conversion)
        self.engine = engine

        # Extract stream information (format, channels, rate) from the engine
        format, channels, rate = self.engine.get_stream_info()

        # Check if the engine doesn't support consuming generators directly
        if not self.engine.can_consume_generators:
            config = AudioConfiguration(
                format,
                channels,
                rate,
                self.output_device_index,
                muted=self.global_muted,
            )

            self.player = StreamPlayer(
                self.engine.queue, config, on_playback_start=self._on_audio_stream_start
            )
        else:
            self.engine.on_playback_start = self._on_audio_stream_start
            self.player = None

        logging.info(f"loaded engine {self.engine.engine_name}")

    def feed(self, text_or_iterator: Union[str, Iterator[str]]):
        """
        Feeds text or an iterator to the stream.

        Args:
            text_or_iterator: Text or iterator to be fed.

        Returns:
            Self instance.
        """
        self.char_iter.add(text_or_iterator)
        return self

    def play_async(
        self,
        fast_sentence_fragment: bool = True,
        fast_sentence_fragment_allsentences: bool = True,
        fast_sentence_fragment_allsentences_multiple: bool = False,
        buffer_threshold_seconds: float = 0.0,
        minimum_sentence_length: int = 10,
        minimum_first_fragment_length: int = 10,
        log_synthesized_text=False,
        reset_generated_text: bool = True,
        output_wavfile: str = None,
        on_sentence_synthesized=None,
        before_sentence_synthesized=None,
        on_audio_chunk=None,
        tokenizer: str = "",
        tokenize_sentences=None,
        language: str = "",
        context_size: int = 12,
        context_size_look_overhead: int = 12,
        muted: bool = False,
        sentence_fragment_delimiters: str = ".?!;:,\n…)]}。-",
        force_first_fragment_after_words=15,
    ):
        """
        Async handling of text to audio synthesis, see play() method.
        """
        if not self.is_playing_flag:
            self.is_playing_flag = True
            # Pass additional parameter to differentiate external call
            args = (
                fast_sentence_fragment,
                fast_sentence_fragment_allsentences,
                fast_sentence_fragment_allsentences_multiple,
                buffer_threshold_seconds,
                minimum_sentence_length,
                minimum_first_fragment_length,
                log_synthesized_text,
                reset_generated_text,
                output_wavfile,
                on_sentence_synthesized,
                before_sentence_synthesized,
                on_audio_chunk,
                tokenizer,
                tokenize_sentences,
                language,
                context_size,
                context_size_look_overhead,
                muted,
                sentence_fragment_delimiters,
                force_first_fragment_after_words,
                True,
            )
            self.play_thread = threading.Thread(target=self.play, args=args)
            self.play_thread.start()
        else:
            logging.warning("play_async() called while already playing audio, skipping")

        # self.play_thread = threading.Thread(target=self.play, args=(fast_sentence_fragment, buffer_threshold_seconds, minimum_sentence_length, minimum_first_fragment_length, log_synthesized_text, reset_generated_text, output_wavfile, on_sentence_synthesized, on_audio_chunk, tokenizer, language, context_size, muted, sentence_fragment_delimiters, force_first_fragment_after_words))
        # self.play_thread.daemon = True
        # self.play_thread.start()

    def play(
        self,
        fast_sentence_fragment: bool = True,
        fast_sentence_fragment_allsentences: bool = False,
        fast_sentence_fragment_allsentences_multiple: bool = False,
        buffer_threshold_seconds: float = 0.0,
        minimum_sentence_length: int = 10,
        minimum_first_fragment_length: int = 10,
        log_synthesized_text=False,
        reset_generated_text: bool = True,
        output_wavfile: str = None,
        on_sentence_synthesized=None,
        before_sentence_synthesized=None,
        on_audio_chunk=None,
        tokenizer: str = "nltk",
        tokenize_sentences=None,
        language: str = "en",
        context_size: int = 12,
        context_size_look_overhead: int = 12,
        muted: bool = False,
        sentence_fragment_delimiters: str = ".?!;:,\n…)]}。-",
        force_first_fragment_after_words=15,
        is_external_call=True,
    ):
        """
        Handles the synthesis of text to audio.
        Plays the audio stream and waits until it is finished playing.
        If the engine can't consume generators, it utilizes a player.

        Args:
        - fast_sentence_fragment: Determines if sentence fragments should be quickly yielded. Useful when a faster response is desired even if a sentence isn't complete.
        - fast_sentence_fragment_allsentences: Fast_sentence_fragment only works on the first sentence. Set this to True if you want to work it on every sentence.
        - fast_sentence_fragment_allsentences_multiple: Can yield multiple sentence fragments, not only a single one.
        - buffer_threshold_seconds (float): Time in seconds for the buffering threshold, influencing the flow and continuity of audio playback. Set to 0 to deactivate. Default is 0.
          - How it Works: The system verifies whether there is more audio content in the buffer than the duration defined by buffer_threshold_seconds. If so, it proceeds to synthesize the next sentence, capitalizing on the remaining audio to maintain smooth delivery. A higher value means more audio is pre-buffered, which minimizes pauses during playback. Adjust this upwards if you encounter interruptions.
          - Helps to decide when to generate more audio based on buffered content.
        - minimum_sentence_length (int): The minimum number of characters a sentence must have. If a sentence is shorter, it will be concatenated with the following one, improving the overall readability. This parameter does not apply to the first sentence fragment, which is governed by `minimum_first_fragment_length`. Default is 10 characters.
        - minimum_first_fragment_length (int): The minimum number of characters required for the first sentence fragment before yielding. Default is 10 characters.
        - log_synthesized_text: If True, logs the synthesized text chunks.
        - reset_generated_text: If True, resets the generated text.
        - output_wavfile: If set, saves the audio to the specified WAV file.
        - on_sentence_synthesized: Callback function that gets called after hen a single sentence fragment was synthesized.
        - before_sentence_synthesized: Callback function that gets called before a single sentence fragment gets synthesized.
        - on_audio_chunk: Callback function that gets called when a single audio chunk is ready.
        - tokenizer: Tokenizer to use for sentence splitting (currently "nltk" and "stanza" are supported).
        - tokenize_sentences (Callable): A function that tokenizes sentences from the input text. You can write your own lightweight tokenizer here if you are unhappy with nltk and stanza. Defaults to None. Takes text as string and should return splitted sentences as list of strings.
        - language: Language to use for sentence splitting.
        - context_size: The number of characters used to establish context for sentence boundary detection. A larger context improves the accuracy of detecting sentence boundaries. Default is 12 characters.
        - muted: If True, disables audio playback via local speakers (in case you want to synthesize to file or process audio chunks). Default is False.
        - sentence_fragment_delimiters (str): A string of characters that are
            considered sentence delimiters. Default is ".?!;:,\n…)]}。-".
        - force_first_fragment_after_words (int): The number of words after
            which the first sentence fragment is forced to be yielded.
            Default is 15 words.
        """
        if self.global_muted:
            muted = True

        if is_external_call:
            if not self.play_lock.acquire(blocking=False):
                logging.warning("play() called while already playing audio, skipping")
                return

        self.is_playing_flag = True

        # Log the start of the stream
        logging.info("stream start")

        tokenizer = tokenizer if tokenizer else self.tokenizer
        language = language if language else self.language

        # Set the stream_running flag to indicate the stream is active
        self.stream_start_time = time.time()
        self.stream_running = True
        abort_event = threading.Event()
        self.abort_events.append(abort_event)

        if self.player:
            self.player.mute(muted)
        elif hasattr(self.engine, "set_muted"):
            self.engine.set_muted(muted)

        self.output_wavfile = output_wavfile
        self.chunk_callback = on_audio_chunk

        if output_wavfile:
            if self._is_engine_mpeg():
                self.wf = open(output_wavfile, "wb")
            else:
                self.wf = wave.open(output_wavfile, "wb")
                _, channels, rate = self.engine.get_stream_info()
                self.wf.setnchannels(channels)
                self.wf.setsampwidth(2)
                self.wf.setframerate(rate)

        # Initialize the generated_text variable
        if reset_generated_text:
            self.generated_text = ""

        # Check if the engine can handle generators directly
        if self.engine.can_consume_generators:
            try:
                # Directly synthesize audio using the character iterator
                self.char_iter.log_characters = self.log_characters

                self.engine.on_audio_chunk = self._on_audio_chunk
                self.engine.synthesize(self.char_iter)

                if self.on_audio_stream_stop:
                    self.on_audio_stream_stop()

            finally:
                # Once done, set the stream running flag to False and log the stream stop
                self.stream_running = False
                logging.info("stream stop")

                # Accumulate the generated text and reset the character iterators
                self.generated_text += self.char_iter.iterated_text

                self._create_iterators()

                if is_external_call:
                    self.is_playing_flag = False
                    self.play_lock.release()
        else:
            try:
                # Start the audio player to handle playback
                self.player.start()
                self.player.on_audio_chunk = self._on_audio_chunk

                # Generate sentences from the characters
                generate_sentences = s2s.generate_sentences(
                    self.thread_safe_char_iter,
                    context_size=context_size,
                    context_size_look_overhead=context_size_look_overhead,
                    minimum_sentence_length=minimum_sentence_length,
                    minimum_first_fragment_length=minimum_first_fragment_length,
                    quick_yield_single_sentence_fragment=fast_sentence_fragment,
                    quick_yield_for_all_sentences=fast_sentence_fragment_allsentences,
                    quick_yield_every_fragment=fast_sentence_fragment_allsentences_multiple,
                    cleanup_text_links=True,
                    cleanup_text_emojis=True,
                    tokenize_sentences=tokenize_sentences,
                    tokenizer=tokenizer,
                    language=language,
                    log_characters=self.log_characters,
                    sentence_fragment_delimiters=sentence_fragment_delimiters,
                    force_first_fragment_after_words=force_first_fragment_after_words,
                )

                # Create the synthesis chunk generator with the given sentences
                chunk_generator = self._synthesis_chunk_generator(
                    generate_sentences, buffer_threshold_seconds, log_synthesized_text
                )

                sentence_queue = queue.Queue()

                def synthesize_worker():
                    while not abort_event.is_set():
                        sentence = sentence_queue.get()
                        if sentence is None:  # Sentinel value to stop the worker
                            break

                        synthesis_successful = False
                        if log_synthesized_text:
                            logging.info(f"synthesizing: {sentence}")

                        while not synthesis_successful:
                            try:
                                if abort_event.is_set():
                                    break

                                if before_sentence_synthesized:
                                    before_sentence_synthesized(sentence)
                                success = self.engine.synthesize(sentence)
                                if success:
                                    if on_sentence_synthesized:
                                        on_sentence_synthesized(sentence)
                                    synthesis_successful = True
                                else:
                                    logging.warning(
                                        f'engine {self.engine.engine_name} failed to synthesize sentence "{sentence}", unknown error'
                                    )

                            except Exception as e:
                                logging.warning(
                                    f'engine {self.engine.engine_name} failed to synthesize sentence "{sentence}" with error: {e}'
                                )
                                tb_str = traceback.format_exc()
                                print(f"Traceback: {tb_str}")
                                print(f"Error: {e}")

                            if not synthesis_successful:
                                if len(self.engines) == 1:
                                    time.sleep(0.2)
                                    logging.warning(
                                        f"engine {self.engine.engine_name} is the only engine available, can't switch to another engine"
                                    )
                                    break
                                else:
                                    logging.warning(
                                        "fallback engine(s) available, switching to next engine"
                                    )
                                    self.engine_index = (self.engine_index + 1) % len(
                                        self.engines
                                    )

                                    self.player.stop()
                                    self.load_engine(self.engines[self.engine_index])
                                    self.player.start()
                                    self.player.on_audio_chunk = self._on_audio_chunk

                        sentence_queue.task_done()

                worker_thread = threading.Thread(target=synthesize_worker)
                worker_thread.start()

                # Iterate through the synthesized chunks and feed them to the engine for audio synthesis
                for sentence in chunk_generator:
                    if abort_event.is_set():
                        break
                    sentence = sentence.strip()
                    if sentence:
                        sentence_queue.put(sentence)
                    else:
                        continue  # Skip empty sentences

                # Signal to the worker to stop
                sentence_queue.put(None)
                worker_thread.join()

            except Exception as e:
                logging.warning(
                    f"error in play() with engine {self.engine.engine_name}: {e}"
                )
                tb_str = traceback.format_exc()
                print(f"Traceback: {tb_str}")
                print(f"Error: {e}")

            finally:
                try:
                    self.player.stop()

                    self.abort_events.remove(abort_event)
                    self.stream_running = False
                    logging.info("stream stop")

                    self.output_wavfile = None
                    self.chunk_callback = None

                finally:
                    if output_wavfile and self.wf:
                        self.wf.close()
                        self.wf = None

            if len(self.char_iter.items) > 0 and self.char_iter.iterated_text == "":
                # new text was feeded while playing audio but after the last character was processed
                # we need to start another play() call (!recursively!)
                self.play(
                    fast_sentence_fragment=fast_sentence_fragment,
                    buffer_threshold_seconds=buffer_threshold_seconds,
                    minimum_sentence_length=minimum_sentence_length,
                    minimum_first_fragment_length=minimum_first_fragment_length,
                    log_synthesized_text=log_synthesized_text,
                    reset_generated_text=False,
                    output_wavfile=output_wavfile,
                    on_sentence_synthesized=on_sentence_synthesized,
                    on_audio_chunk=on_audio_chunk,
                    tokenizer=tokenizer,
                    language=language,
                    context_size=context_size,
                    muted=muted,
                    sentence_fragment_delimiters=sentence_fragment_delimiters,
                    force_first_fragment_after_words=force_first_fragment_after_words,
                    is_external_call=False,
                )

            if is_external_call:
                if self.on_audio_stream_stop:
                    self.on_audio_stream_stop()

                self.is_playing_flag = False
                self.play_lock.release()

    def pause(self):
        """
        Pauses playback of the synthesized audio stream (won't work properly with elevenlabs).
        """
        if self.is_playing():
            logging.info("stream pause")
            if self.engine.can_consume_generators:
                self.engine.pause()
            else:
                self.player.pause()

    def resume(self):
        """
        Resumes a previously paused playback of the synthesized audio stream
        - won't work properly with elevenlabs
        """
        if self.is_playing():
            logging.info("stream resume")
            if self.engine.can_consume_generators:
                self.engine.resume()
            else:
                self.player.resume()

    def stop(self):
        """
        Stops the playback of the synthesized audio stream immediately.
        """

        for abort_event in self.abort_events:
            abort_event.set()

        if self.is_playing():
            self.char_iter.stop()
            if self.engine.can_consume_generators:
                if self.engine.stop():
                    self.stream_running = False
            else:
                self.player.stop(immediate=True)
                self.stream_running = False

        if self.play_thread is not None:
            if self.play_thread.is_alive():
                self.play_thread.join()
            self.play_thread = None

        self._create_iterators()

    def text(self):
        """
        Retrieves the text that has been fed into the stream.

        Returns:
            The accumulated text.
        """
        if self.generated_text:
            return self.generated_text
        return self.thread_safe_char_iter.accumulated_text()

    def is_playing(self):
        """
        Checks if the stream is currently playing.

        Returns:
            Boolean indicating if the stream is playing.
        """
        return self.stream_running

    def _on_audio_stream_start(self):
        """
        Handles the start of the audio stream.

        This method is called when the audio stream starts. It calculates and logs the latency from the stream's start time to the time when the first chunk of audio is received. If a callback for handling the start of the audio stream is set (on_audio_stream_start), it is executed.

        No parameters or returns.
        """
        latency = time.time() - self.stream_start_time
        logging.info(f"Audio stream start, latency to first chunk: {latency:.2f}s")

        if self.on_audio_stream_start:
            self.on_audio_stream_start()

    def _on_audio_chunk(self, chunk):
        """
        Postprocessing of single chunks of audio data.
        This method is called for each chunk of audio data processed. It first determines the audio stream format.
        If the format is `pyaudio.paFloat32`, we convert to paInt16.

        Args:
            chunk (bytes): The audio data chunk to be processed.
        """
        format, _, _ = self.engine.get_stream_info()

        if format == pyaudio.paFloat32:
            audio_data = np.frombuffer(chunk, dtype=np.float32)
            audio_data = np.int16(audio_data * 32767)
            chunk = audio_data.tobytes()

        if self.output_wavfile and self.wf:
            if self._is_engine_mpeg():
                self.wf.write(chunk)
            else:
                self.wf.writeframes(chunk)

        if self.chunk_callback:
            self.chunk_callback(chunk)

    def _on_last_character(self):
        """
        This method is invoked when the last character of the text stream has been processed.
        It logs information and triggers a callback, if defined.
        """

        # If an on_text_stream_stop callback is defined, invoke it to signal the end of the text stream
        if self.on_text_stream_stop:
            self.on_text_stream_stop()

        # If log_characters flag is True, print a new line for better log readability
        if self.log_characters:
            print()

        self._create_iterators()

    def _create_iterators(self):
        """
        Creates iterators required for text-to-audio streaming.

        This method initializes two types of iterators:

        1. `CharIterator`: Responsible for managing individual characters during the streaming process.
        - It takes callbacks for events like when a character is processed (`on_character`), when the first text chunk is encountered (`on_first_text_chunk`), and when the last text chunk is encountered (`on_last_text_chunk`).

        2. `AccumulatingThreadSafeGenerator`: A thread-safe wrapper around `CharIterator`.
        - Ensures that the character iterator can be safely accessed from multiple threads.
        """

        # Create a CharIterator instance for managing individual characters
        self.char_iter = CharIterator(
            on_character=self._on_character,
            on_first_text_chunk=self.on_text_stream_start,
            on_last_text_chunk=self._on_last_character,
        )

        # Create a thread-safe version of the char iterator
        self.thread_safe_char_iter = AccumulatingThreadSafeGenerator(self.char_iter)

    def _on_character(self, char: str):
        """
        This method is called for each character that is processed in the text stream.
        It accumulates the characters and invokes a callback.

        Args:
            char (str): The character currently being processed.
        """
        # If an on_character callback is defined, invoke it for the current character
        if self.on_character:
            self.on_character(char)

        self.generated_text += char

    def _is_engine_mpeg(self):
        """
        Checks if the engine is an MPEG engine.

        Returns:
            Boolean indicating if the engine is an MPEG engine.
        """
        format, channel, rate = self.engine.get_stream_info()
        return format == pyaudio.paCustomFormat and channel == -1 and rate == -1

    def _synthesis_chunk_generator(
        self,
        generator: Iterator[str],
        buffer_threshold_seconds: float = 2.0,
        log_synthesis_chunks: bool = False,
    ) -> Iterator[str]:
        """
        Generates synthesis chunks based on buffered audio length.

        The function buffers chunks of synthesis until the buffered audio seconds fall below the provided threshold.
        Once the threshold is crossed, the buffered synthesis chunk is yielded.

        Args:
            generator: Input iterator that provides chunks for synthesis.
            buffer_threshold_seconds: Time in seconds to specify how long audio data should be buffered before yielding the synthesis chunk.
            log_synthesis_chunks: Boolean flag that, if set to True, logs the synthesis chunks to the logging system.

        Returns:
            Iterator of synthesis chunks.
        """

        # Initializes an empty string to accumulate chunks of synthesis
        synthesis_chunk = ""

        # Iterates over each chunk from the provided generator
        for chunk in generator:
            # Fetch the total seconds of buffered audio
            buffered_audio_seconds = self.player.get_buffered_seconds()

            # Append the current chunk (and a space) to the accumulated synthesis_chunk
            synthesis_chunk += chunk + " "

            # Check if the buffered audio is below the specified threshold
            if (
                buffered_audio_seconds < buffer_threshold_seconds
                or buffer_threshold_seconds <= 0
            ):
                # If the log_synthesis_chunks flag is True, log the current synthesis_chunk
                if log_synthesis_chunks:
                    logging.info(
                        f'-- ["{synthesis_chunk}"], buffered {buffered_audio_seconds:1f}s'
                    )

                # Yield the current synthesis_chunk and reset it for the next set of accumulations
                yield synthesis_chunk
                synthesis_chunk = ""

            else:
                logging.info(
                    f"summing up chunks because buffer {buffered_audio_seconds:.1f} > threshold ({buffer_threshold_seconds:.1f}s)"
                )

        # After iterating over all chunks, check if there's any remaining data in synthesis_chunk
        if synthesis_chunk:
            # If the log_synthesis_chunks flag is True, log the remaining synthesis_chunk
            if log_synthesis_chunks:
                logging.info(
                    f'-- ["{synthesis_chunk}"], buffered {buffered_audio_seconds:.1f}s'
                )

            # Yield the remaining synthesis_chunk
            yield synthesis_chunk



from .base_engine import BaseEngine
from pydub.utils import mediainfo
from pydub import AudioSegment
from typing import Union
import tempfile
import pyaudio
import pyttsx3
import wave
import os

SYNTHESIS_FILE = "system_speech_synthesis.wav"


class SystemVoice:
    def __init__(self, name, id):
        self.name = name
        self.id = id

    def __repr__(self):
        return self.name
        # return f"<Voice(name={self.name})>"


class SystemEngine(BaseEngine):
    def __init__(self, voice: str = "Zira", print_installed_voices: bool = False):
        """
        Initializes a system realtime text to speech engine object.

        Args:
            voice (str, optional): Voice name. Defaults to "Zira".
            print_installed_voices (bool, optional): Indicates if the list of installed voices should be printed. Defaults to False.
        """

        self.engine = pyttsx3.init()
        self.set_voice(voice)
        temp_file_path = tempfile.gettempdir()
        self.file_path = os.path.join(temp_file_path, SYNTHESIS_FILE)

        if print_installed_voices:
            print(self.get_voices())

    def post_init(self):
        self.engine_name = "system"

    def get_stream_info(self):
        """
        Returns the PyAudio stream configuration information suitable for System Engine.

        Returns:
            tuple: A tuple containing the audio format, number of channels, and the sample rate.
                  - Format (int): The format of the audio stream. pyaudio.paInt16 represents 16-bit integers.
                  - Channels (int): The number of audio channels. 1 represents mono audio.
                  - Sample Rate (int): The sample rate of the audio in Hz. 16000 represents 16kHz sample rate.
        """
        return pyaudio.paInt16, 1, 22050

    def synthesize(self, text: str) -> bool:
        """
        Synthesizes text to audio stream.

        Args:
            text (str): Text to synthesize.
        """

        self.engine.save_to_file(text, self.file_path)
        self.engine.runAndWait()

        # Get media info of the file
        info = mediainfo(self.file_path)

        # Check if the file format is AIFF and convert to WAV if necessary
        if info["format_name"] == "aiff":
            audio = AudioSegment.from_file(self.file_path, format="aiff")
            audio.export(self.file_path, format="wav")

        # Now open the WAV file
        with wave.open(self.file_path, "rb") as wf:
            audio_data = wf.readframes(wf.getnframes())
            self.queue.put(audio_data)
            return True

        # Return False if the process failed
        return False

    def get_voices(self):
        """
        Retrieves the voices available in the underlying system's speech engine.

        This method fetches the list of available voices from the system's speech engine
        (represented by the `engine` attribute of the instance) and wraps each voice's information
        in a `SystemVoice` object.

        Returns:
            list[SystemVoice]: A list containing SystemVoice objects representing each available voice.
                            Each SystemVoice object encapsulates information like the name and ID of the voice.

        Note:
            Ensure that the `engine` attribute of the instance is properly initialized before calling this method.
        """
        voice_objects = []
        voices = self.engine.getProperty("voices")
        for voice in voices:
            voice_object = SystemVoice(voice.name, voice.id)
            voice_objects.append(voice_object)
        return voice_objects

    def set_voice(self, voice: Union[str, SystemVoice]):
        """
        Sets the voice to be used for speech synthesis.

        Args:
            voice (Union[str, SystemVoice]): The voice to be used for speech synthesis.
        """
        if isinstance(voice, SystemVoice):
            self.engine.setProperty("voice", voice.id)
        else:
            installed_voices = self.engine.getProperty("voices")
            if voice is not None:
                for installed_voice in installed_voices:
                    if voice in installed_voice.name:
                        self.engine.setProperty("voice", installed_voice.id)

    def set_voice_parameters(self, **voice_parameters):
        """
        Sets the voice parameters to be used for speech synthesis.

        Args:
            **voice_parameters: The voice parameters to be used for speech synthesis.
        """
        for parameter, value in voice_parameters.items():
            self.engine.setProperty(parameter, value)



from .base_engine import BaseEngine
from openai import OpenAI
from typing import Union
import pyaudio


class OpenAIVoice:
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return f"{self.name}"


class OpenAIEngine(BaseEngine):
    def __init__(self, model: str = "tts-1", voice: str = "nova"):
        """
        Initializes a openai realtime text to speech engine object.

        Args:
            model (str, optional): The model to use for speech synthesis.
              Available models: tts-1-hd and tts-1. Defaults to "tts-1".
            voice (str, optional): The voice to use for speech synthesis.
              Available voices: alloy, echo, fable, onyx, nova, and shimmer.
              Defaults to "nova".
        """

        self.voices = ["alloy", "echo", "fable", "onyx", "nova", "shimmer"]
        self.model = model
        self.voice = voice
        self.client = OpenAI()

    def post_init(self):
        self.engine_name = "openai"

    def get_stream_info(self):
        """
        Returns the PyAudio stream configuration information suitable
        for OpenAI Engine.

        Returns:
            tuple: A tuple containing the audio format, number of channels,
              and the sample rate.
                  - Format (int): The format of the audio stream.
                    pyaudio.paInt16 represents 16-bit integers.
                  - Channels (int): The number of audio channels.
                    1 represents mono audio.
                  - Sample Rate (int): The sample rate of the audio in Hz.
                    16000 represents 16kHz sample rate.
        """
        return pyaudio.paCustomFormat, 1, 22050

    def synthesize(self, text: str) -> bool:
        """
        Synthesizes text to audio stream.

        Args:
            text (str): Text to synthesize.
        """
        response = self.client.audio.speech.create(
            model=self.model,
            voice=self.voice,
            input=text,
        )

        for data in response.iter_bytes():
            # we write the mp3 raw data into the queue
            # and let stream player handle the rest
            self.queue.put(data)

        return True

    def get_voices(self):
        """
        Retrieves the installed voices available for the Coqui TTS engine.
        """
        voice_objects = []
        for voice in self.voices:
            voice_objects.append(OpenAIVoice(voice))
        return voice_objects

    def set_voice(self, voice: Union[str, object]):
        """
        Sets the voice to be used for speech synthesis.

        Args:
            voice (Union[str, SystemVoice]): The voice to be used for speech
              synthesis.
        """
        if isinstance(voice, OpenAIVoice):
            self.voice = voice.name
        else:
            installed_voices = self.get_voices()
            for installed_voice in installed_voices:
                if voice in installed_voice.name:
                    self.voice = installed_voice.name

    def set_voice_parameters(self, **voice_parameters):
        """
        Sets the voice parameters to be used for speech synthesis.

        Args:
            **voice_parameters: The voice parameters to be used for speech
              synthesis.
        """
        pass



import setuptools

# Read the contents of README.md
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


# Read requirements.txt and parse it
def parse_requirements(filename):
    requirements = {}
    with open(filename, "r") as f:
        for line in f:
            line = line.strip()
            if line and not line.startswith("#"):
                package = line.split("==")[0]
                requirements[package] = line
    return requirements


requirements = parse_requirements("requirements.txt")

print("Requirements:", requirements)

# Define all requirements
all_requirements = list(requirements.values())

# Define base requirements
base_requirements = [
    requirements["stream2sentence"],
    requirements["pydub"],
    requirements["pyaudio"],
]

# Define subsets of requirements for each engine
extras_require = {
    "minimal": base_requirements,
    "all": base_requirements
    + [requirements["pyttsx3"]]
    + [requirements["azure-cognitiveservices-speech"]]
    + [requirements["elevenlabs"]]
    + [requirements["openai"]]
    + [requirements["gtts"]]
    + [requirements["coqui_tts"]],
    "system": base_requirements + [requirements["pyttsx3"]],
    "azure": base_requirements + [requirements["azure-cognitiveservices-speech"]],
    "elevenlabs": base_requirements + [requirements["elevenlabs"]],
    "openai": base_requirements + [requirements["openai"]],
    "gtts": base_requirements + [requirements["gtts"]],
    "coqui": base_requirements + [requirements["coqui_tts"]],
    "jp": base_requirements + ["mecab-python3==1.0.6", "unidic-lite==1.0.8", "cutlet"],
    "ko": base_requirements + ["hangul_romanize"],
}

setuptools.setup(
    name="RealTimeTTS",
    version="0.4.10",
    author="Kolja Beigel",
    author_email="kolja.beigel@web.de",
    description="Stream text into audio with an easy-to-use, highly configurable library delivering voice output with minimal latency.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/KoljaB/RealTimeTTS",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9, <3.13",
    install_requires=base_requirements,
    extras_require=extras_require,
    package_data={"RealtimeTTS": ["engines/*.json"]},
    include_package_data=True,
    keywords="real-time, text-to-speech, TTS, streaming, audio, voice, synthesis, sentence-segmentation, low-latency, character-streaming, dynamic feedback, audio-output, text-input, TTS-engine, audio-playback, stream-player, sentence-fragment, audio-feedback, interactive, python",
)



from RealtimeTTS import TextToAudioStream, AzureEngine
import os
import openai  # pip install openai
import time

openai.api_key = os.environ.get("OPENAI_API_KEY")
engine = AzureEngine(
    os.environ.get("AZURE_SPEECH_KEY"), os.environ.get("AZURE_SPEECH_REGION")
)


def generate(prompt):
    for chunk in openai.ChatCompletion.create(
        model="gpt-3.5-turbo",
        messages=[{"role": "user", "content": prompt}],
        stream=True,
    ):
        if text_chunk := chunk["choices"][0]["delta"].get("content"):
            yield text_chunk


def text_start():
    print("[TEXT START]", end="", flush=True)


def text_stop():
    print("[TEXT STOP]", end="", flush=True)


def audio_start():
    print("[AUDIO START]", end="", flush=True)


def audio_stop():
    print("[AUDIO STOP]", end="", flush=True)


stream = TextToAudioStream(
    engine,
    on_text_stream_start=text_start,
    on_text_stream_stop=text_stop,
    on_audio_stream_start=audio_start,
    on_audio_stream_stop=audio_stop,
    log_characters=True,
)

text_stream = generate("A very short two-sentence relaxing speech.")
stream.feed(text_stream)
stream.play_async()

while stream.is_playing():
    time.sleep(0.1)



import json
import logging
import threading
import asyncio
import argparse
from queue import Queue, Empty
import websockets
from xtts_rvc_synthesizer import XTTSRVCSynthesizer
from concurrent.futures import ThreadPoolExecutor

# Use thread-safe Queues for audio chunks and control messages
audio_queue = Queue()
control_queue = Queue()

# Set to store audio WebSocket connections
audio_connections = set()

# Event to signal the threads to stop
stop_event = threading.Event()

class TTSThread(threading.Thread):
    def __init__(self, xtts_model, xtts_voice, rvc_model, use_logging):
        super().__init__()
        self.xtts_model = xtts_model
        self.xtts_voice = xtts_voice
        self.rvc_model = rvc_model
        self.use_logging = use_logging
        self.tts = None
        self.rvc = True
        self.ready = threading.Event()
        self.logger = logging.getLogger('TTSThread')
        self.current_synthesis_task = None
        self.synthesis_executor = ThreadPoolExecutor(max_workers=1)

    def run(self):
        self.logger.info("Initializing TTS...")
        self.tts = XTTSRVCSynthesizer(
            xtts_model=self.xtts_model,
            xtts_voice=self.xtts_voice,
            rvc_model=self.rvc_model,
            rvc_sample_rate=40000,
            use_logging=self.use_logging,
            on_audio_chunk=self.on_audio_chunk
        )

        self.logger.info("TTS Server ready")
        print("TTS Server ready...")
        self.ready.set()  # Signal that TTS is ready
       
        while not stop_event.is_set():
            try:
                data = control_queue.get(timeout=0.1)
                self.logger.debug(f"Received control data: {data}")
                if data["type"] == "text":
                    if data["rvc"] and not self.rvc:
                        self.logger.info("Enabling RVC")
                        self.rvc = True
                        self.tts.enable_rvc(True)
                    elif not data["rvc"] and self.rvc:
                        self.logger.info("Disabling RVC")
                        self.rvc = False
                        self.tts.enable_rvc(False)

                    self.logger.info(f"Pushing text to TTS: {data['content'][:50]}")
                    self.tts.push_text(data["content"])
                elif data["type"] == "synthesize":
                    self.logger.info("Starting synthesis")
                    self.current_synthesis_task = self.synthesis_executor.submit(self.tts.synthesize)
                    # self.tts.synthesize()
                elif data["type"] == "cancel":
                    if self.current_synthesis_task:
                        self.current_synthesis_task.cancel()                    
                    self.logger.info("Cancelling synthesis")
                    self.tts.stop()
                    self.clear_audio_queue()
                elif data["type"] == "new_connection":
                    self.logger.info("New connection detected, stopping synthesis and clearing audio queue")
                    if self.current_synthesis_task:
                        self.current_synthesis_task.cancel()                    
                    self.tts.stop()
                    self.clear_audio_queue()                    
            except Empty:
                continue

    def stop(self):
        self.synthesis_executor.shutdown(wait=False)                

    def on_audio_chunk(self, chunk):
        self.logger.debug(f"Received audio chunk of size {len(chunk)}")
        audio_queue.put(chunk)

    def clear_audio_queue(self):
        self.logger.info("Clearing audio queue")
        # while not audio_queue.empty():
        while audio_queue.qsize() > 0:
            audio_queue.get_nowait()
        self.logger.info("Audio queue cleared after cancel")

async def process_audio_queue():
    logger = logging.getLogger('AudioProcessor')
    while True:
        try:
            chunk = audio_queue.get_nowait()
            logger.debug(f"Processing audio chunk of size {len(chunk)}")
            await broadcast_audio_chunk(chunk)
        except Empty:
            await asyncio.sleep(0.01)

async def broadcast_audio_chunk(chunk):
    logger = logging.getLogger('AudioBroadcaster')
    logger.debug(f"Broadcasting audio chunk to {len(audio_connections)} connections")
    for conn in list(audio_connections):
        try:
            await conn.send(chunk)
        except websockets.exceptions.ConnectionClosed:
            logger.info(f"Removing closed connection {conn}")
            audio_connections.remove(conn)

async def control_handler(websocket, path):
    logger = logging.getLogger('ControlHandler')
    try:
        logger.info(f"New control connection from {websocket.remote_address}")
        # Signal new connection to stop synthesis and clear audio queue
        control_queue.put({"type": "new_connection"})
        async for message in websocket:
            data = json.loads(message)
            logger.debug(f"Received control message: {data}")
            control_queue.put(data)
            await websocket.send(json.dumps({"type": f"{data['type']}_received"}))
    except websockets.exceptions.ConnectionClosed:
        logger.info(f"Control WebSocket connection closed from {websocket.remote_address}")

async def audio_handler(websocket, path):
    logger = logging.getLogger('AudioHandler')
    try:
        logger.info(f"New audio connection from {websocket.remote_address}")
        audio_connections.add(websocket)
        await websocket.wait_closed()
    finally:
        logger.info(f"Audio WebSocket connection closed from {websocket.remote_address}")
        audio_connections.remove(websocket)

async def main_async(args):
    logger = logging.getLogger('Main')
    logger.info("Starting TTS Server")

    # Start the TTS thread
    tts_thread = TTSThread(args.xtts_model, args.xtts_voice, args.rvc_model, args.use_logging)
    tts_thread.start()

    # Wait for TTS to be ready
    logger.info("Waiting for TTS to be ready")
    await asyncio.get_event_loop().run_in_executor(None, tts_thread.ready.wait)

    # Start the audio processing task
    logger.info("Starting audio processing task")
    audio_task = asyncio.create_task(process_audio_queue())

    logger.info(f"Starting control server on {args.host}:{args.control_port}")
    control_server = await websockets.serve(control_handler, args.host, args.control_port)
    
    logger.info(f"Starting audio server on {args.host}:{args.audio_port}")
    audio_server = await websockets.serve(audio_handler, args.host, args.audio_port)
    
    print(f"Server CONTROL listening on ws://{args.host}:{args.control_port}")
    print(f"Server AUDIO listening on ws://{args.host}:{args.audio_port}")
    
    try:
        await asyncio.gather(control_server.wait_closed(), audio_server.wait_closed(), audio_task)
    finally:
        logger.info("Shutting down TTS Server")
        stop_event.set()
        tts_thread.join()

def main():
    parser = argparse.ArgumentParser(description="TTS Server with WebSocket interface")
    parser.add_argument("--xtts-model", default="D:/Data/Models/xtts/v2.0.2", help="Path to XTTS model")
    parser.add_argument("--xtts-voice", default="vanessa.wav", help="XTTS voice file")
    parser.add_argument("--rvc-model", default="models/rvc/Lasinya", help="Path to RVC model")
    parser.add_argument("--host", default="localhost", help="Host to bind the server to")
    parser.add_argument("--control-port", type=int, default=8000, help="Port for control WebSocket")
    parser.add_argument("--audio-port", type=int, default=8001, help="Port for audio WebSocket")
    parser.add_argument("--use-logging", action="store_true", help="Enable detailed logging")
    parser.add_argument("--log-level", default="INFO", choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], help="Set the logging level")
    
    args = parser.parse_args()
    
    logging.basicConfig(
        level=getattr(logging, args.log_level),
        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
        datefmt='%Y-%m-%d %H:%M:%S'
    )
    
    asyncio.run(main_async(args))

if __name__ == "__main__":
    main()